"""
This module contains information about the way the cluster expansion is implemented in the package.
"""
import functools
import operator
import warnings

import numpy as np
from pycce.sm import _smc

try:
    from mpi4py import MPI

    mpiop = {'imul': MPI.PROD,
             'mul': MPI.PROD,
             'prod': MPI.PROD,
             'iadd': MPI.SUM,
             'add': MPI.SUM,
             'sum': MPI.SUM
             }

except ImportError:
    mpiop = None


def cluster_expansion_decorator(_func=None, *,
                                result_operator=operator.imul,
                                contribution_operator=operator.ipow,
                                removal_operator=operator.itruediv,
                                addition_operator=np.prod):
    """
    Decorator for creating cluster correlation expansion of the method of ``RunObject`` class.

    Args:
        _func (func): Function to expand.
        result_operator (func):
            Operator which will combine the result of expansion (default: operator.imul).
        contribution_operator (func):
            Operator which will combine multiple contributions
            of the same cluster (default: operator.ipow) in the optimized approach.
        result_operator (func):
            Operator which will combine the result of expansion (default: operator.imul).
        removal_operator (func):
            Operator which will remove subcluster contribution from the given cluster contribution.
            First argument cluster contribution, second - subcluster contribution (default: operator.itruediv).
        addition_operator (func):
            Group operation which will combine contributions from the different clusters into one
            contribution (default: np.prod).

    Returns:
        func: Expanded function.
    """

    def inner_cluster_expansion_decorator(function):

        @functools.wraps(function)
        def cluster_expansion(self, *arg, **kwarg):

            if self.direct:
                return direct_approach(function, self, *arg,
                                       result_operator=result_operator,
                                       removal_operator=removal_operator,
                                       addition_operator=addition_operator,
                                       **kwarg)
            else:
                return optimized_approach(function, self, *arg,
                                          result_operator=result_operator,
                                          contribution_operator=contribution_operator,
                                          **kwarg)

        return cluster_expansion

    if _func is None:
        return inner_cluster_expansion_decorator
    else:
        return inner_cluster_expansion_decorator(_func)


def optimized_approach(function, self, *arg,
                       result_operator=operator.imul,
                       contribution_operator=operator.ipow,
                       **kwarg):
    """
    Optimized approach to compute cluster correlation expansion.

    Args:
        function (func): Function to expand.
        self (RunObject): Object whose method is expanded.
        *arg: list of positional arguments of the expanded function.
        result_operator (func):
            Operator which will combine the result of expansion (default: operator.imul).
        contribution_operator (func):
            Operator which will combine multiple contributions
            of the same cluster (default: operator.ipow).
        **kwarg: Dictionary containing all keyword arguments of the expanded function.

    Returns:
        func: Expanded function.
    """
    subclusters = self.clusters
    revorders = sorted(subclusters)[::-1]
    norders = len(revorders)

    if self.parallel:
        try:
            from mpi4py import MPI
        except ImportError:
            warnings.warn('Parallel failed: mpi4py is not found. Running serial.')
            self.parallel = False

    if self.parallel:
        comm = MPI.COMM_WORLD

        size = comm.Get_size()
        rank = comm.Get_rank()
    else:
        rank = 0

    # If there is only one set of indexes for only one order,
    # Then for this subcluster nelements < maximum CCE order
    if norders == 1 and subclusters[revorders[0]].shape[0] == 1:
        verticles = subclusters[revorders[0]][0]
        return function(verticles, *arg, **kwarg)

    result = 1
    result = contribution_operator(result, 0)
    # The Highest possible L will have all powers of 1
    power = {}
    # Number of visited orders from highest to lowest
    visited = 0
    for order in revorders:
        nclusters = subclusters[order].shape[0]
        current_power = np.ones(nclusters, dtype=np.int32)
        # indexes of the cluster of size order are stored in v
        if self.parallel:
            remainder = nclusters % size
            add = int(rank < remainder)
            each = nclusters // size
            block = each + add
            start = rank * each + rank if rank < remainder else rank * each + remainder
        else:
            start = 0
            block = nclusters

        for index in range(start, start + block):

            v = subclusters[order][index]
            # First, find the correct power. Iterate over all higher orders
            for higherorder in revorders[:visited]:
                # np.isin gives bool array of shape subclusters[higherorder],
                # which is np.array of
                # indexes of subclusters with order = higherorder.
                # Entries are True if value is
                # present in v and False if values are not present in v.
                # Sum bool entries in inside cluster,
                # if the sum equal to size of v,
                # then v is inside the given subcluster.
                # containv is 1D bool array with values of i-element True
                # if i-subcluster of
                # subclusters[higherorder] contains v
                containv = np.count_nonzero(
                    np.isin(subclusters[higherorder], v), axis=1) == v.size

                # Power of cluster v is decreased by sum of powers of all the higher orders,
                # As all of them have to be divided by v
                current_power[index] -= np.sum(power[higherorder][containv], dtype=np.int32)

            if current_power[index] == 0:
                vcalc = contribution_operator(1, 0)
            else:
                vcalc = function(v, *arg, **kwarg)
                vcalc = contribution_operator(vcalc, current_power[index])

            result = result_operator(result, vcalc)

        if self.parallel:
            buffer = np.empty(current_power.shape, dtype=np.int32)
            comm.Allreduce(current_power, buffer, MPI.SUM)
            current_power = buffer - size + 1

        power[order] = current_power

        visited += 1
        # print('Computed {} of order {} for {} clusters'.format(
        #     function.__name__, order, subclusters[order].shape[0]))
    _smc.clear()

    if self.parallel:
        if rank == 0:
            result_shape = result.shape
        else:
            result_shape = None
        result_shape = comm.bcast(result_shape, root=0)
        if np.asarray(result).shape != result_shape:
            result = np.ones(result_shape, dtype=np.complex128)
            result = contribution_operator(result, 0)

        root_result = np.zeros(result_shape, dtype=np.complex128)
        comm.Allreduce(result.astype(np.complex128), root_result, mpiop[result_operator.__name__])

    else:
        root_result = result

    return root_result


def direct_approach(function, self, *arg,
                    result_operator=operator.imul,
                    removal_operator=operator.itruediv,
                    addition_operator=np.prod,
                    **kwarg):
    """
    Direct approach to compute cluster correlation expansion.

    Args:
        function (func): Function to expand.
        self (RunObject): Object whose method is expanded.
        result_operator (func):
            Operator which will combine the result of expansion (default: operator.imul).
        removal_operator (func):
            Operator which will remove subcluster contribution from the given cluster contribution.
            First argument cluster contribution, second - subcluster contribution (default: operator.itruediv).
        addition_operator (func):
            Group operation which will combine contributions from the different clusters into one
            contribution (default: np.prod).
        **kwarg: Dictionary containing all keyword arguments of the expanded function.

    Returns:
        func: Expanded method.

    """
    subclusters = self.clusters

    if self.parallel:
        try:
            from mpi4py import MPI
        except ImportError:
            warnings.warn('Parallel failed: mpi4py is not found. Running serial')
            self.parallel = False
            MPI = None

    orders = sorted(subclusters)
    norders = len(orders)

    if self.parallel:
        comm = MPI.COMM_WORLD

        size = comm.Get_size()
        rank = comm.Get_rank()
    else:
        rank = 0
        comm = None
    # print(dms_zero.mask)
    # If there is only one set of indexes for only one order,
    # Then for this subcluster nelements < maximum CCE order
    if norders == 1 and subclusters[orders[0]].shape[0] == 1:
        verticles = subclusters[orders[0]][0]

        return function(verticles, *arg, **kwarg)

        # print(zero_power)
    # The Highest possible L will have all powers of 1
    result_tilda = {}
    visited = 0
    result = 1 - result_operator(1, 0)

    for order in orders:
        current_order = []
        # indexes of the cluster of size order are stored in v
        nclusters = subclusters[order].shape[0]

        if self.parallel:
            remainder = nclusters % size
            add = int(rank < remainder)
            each = nclusters // size
            block = each + add
            start = rank * each + rank if rank < remainder else rank * each + remainder
        else:
            start = 0
            block = nclusters

        for index in range(start, start + block):

            v = subclusters[order][index]
            vcalc = function(v, *arg, **kwarg)

            for lowerorder in orders[:visited]:
                contained_in_v = np.all(np.isin(subclusters[lowerorder], v), axis=1)
                lower_vcalc = addition_operator(result_tilda[lowerorder][contained_in_v], axis=0)
                vcalc = removal_operator(vcalc, lower_vcalc)

            current_order.append(vcalc)
        current_order = np.array(current_order, copy=False)

        if self.parallel:
            comm.Barrier()
            result_shape = vcalc.shape if rank == 0 else None
            result_shape = comm.bcast(result_shape, root=0)

            chunk = np.zeros((nclusters, *result_shape), dtype=np.complex128)
            chunk[start:start + block] = current_order.reshape(block, *result_shape)

            currrent_buffer = np.zeros((nclusters, *result_shape), dtype=np.complex128)
            comm.Allreduce(chunk, currrent_buffer, MPI.SUM)
            current_order = currrent_buffer

        result_tilda[order] = current_order
        visited += 1

    for o in orders:
        result = result_operator(result, addition_operator(result_tilda[o], axis=0))

    return result


def interlaced_decorator(_func=None, *,
                         result_operator=operator.imul,
                         contribution_operator=operator.ipow):
    """
    Decorator for creating interlaced cluster correlation expansion of the method of ``RunObject`` class.

    Args:
        _func (func): Function to expand.
        result_operator (func):
            Operator which will combine the result of expansion (default: operator.imul).
        contribution_operator (func):
            Operator which will combine multiple contributions
            of the same cluster (default: operator.ipow) in the optimized approach.

    Returns:
        func: Expanded method.

    """

    def inner_interlaced_decorator(function):

        @functools.wraps(function)
        def cluster_expansion(self, *arg, **kwarg):

            subclusters = self.clusters
            revorders = sorted(subclusters)[::-1]
            norders = len(revorders)

            if self.parallel:
                try:
                    from mpi4py import MPI
                except ImportError:
                    warnings.warn('Parallel failed: mpi4py is not found. Running serial.')
                    self.parallel = False

            if self.parallel:
                comm = MPI.COMM_WORLD

                size = comm.Get_size()
                rank = comm.Get_rank()
            else:
                rank = 0

            # If there is only one set of indexes for only one order,
            # Then for this subcluster nelements < maximum CCE order
            if norders == 1 and subclusters[revorders[0]].shape[0] == 1:
                verticles = subclusters[revorders[0]][0]
                return function(self, verticles, *arg, **kwarg)

            result = 1
            result = contribution_operator(result, 0)
            # The Highest possible L will have all powers of 1
            power = {}
            # Number of visited orders from highest to lowest
            visited = 0
            for order in revorders:
                nclusters = subclusters[order].shape[0]
                current_power = np.ones(nclusters, dtype=np.int32)
                # indexes of the cluster of size order are stored in v
                if self.parallel:
                    remainder = nclusters % size
                    add = int(rank < remainder)
                    each = nclusters // size
                    block = each + add
                    start = rank * each + rank if rank < remainder else rank * each + remainder
                else:
                    start = 0
                    block = nclusters

                for index in range(start, start + block):

                    v = subclusters[order][index]
                    supercluster = []

                    for higherorder in revorders[:visited]:
                        containv = np.count_nonzero(
                            np.isin(subclusters[higherorder], v), axis=1) == v.size

                        supercluster.append(subclusters[higherorder][containv].ravel())
                        current_power[index] -= np.sum(power[higherorder][containv], dtype=np.int32)

                    try:
                        supercluster = np.unique(np.concatenate(supercluster))
                    except ValueError:
                        supercluster = v
                    if not supercluster.size:
                        supercluster = v

                    vcalc = function(v, supercluster, *arg, **kwarg)
                    vcalc = contribution_operator(vcalc, current_power[index])

                    result = result_operator(result, vcalc)

                if self.parallel:
                    buffer = np.empty(current_power.shape, dtype=np.int32)
                    comm.Allreduce(current_power, buffer, MPI.SUM)
                    current_power = buffer - size + 1

                power[order] = current_power

                visited += 1
                # print('Computed {} of order {} for {} clusters'.format(
                #     function.__name__, order, subclusters[order].shape[0]))
            _smc.clear()

            if self.parallel:
                if rank == 0:
                    result_shape = result.shape
                else:
                    result_shape = None
                result_shape = comm.bcast(result_shape, root=0)
                if np.asarray(result).shape != result_shape:
                    result = np.ones(result_shape, dtype=np.complex128)
                    result = contribution_operator(result, 0)

                root_result = np.zeros(result_shape, dtype=np.complex128)
                comm.Allreduce(result.astype(np.complex128), root_result, mpiop[result_operator.__name__])

            else:
                root_result = result

            return root_result

        return cluster_expansion

    if _func is None:
        return inner_interlaced_decorator
    else:
        return inner_interlaced_decorator(_func)



import operator

import numpy as np
from numba import jit
from pycce.constants import PI2
from pycce.h.functions import external_spins_field
from pycce.run.clusters import cluster_expansion_decorator, interlaced_decorator
from pycce.run.mc import monte_carlo_method_decorator
from pycce.run.pulses import Sequence
from pycce.sm import numba_gen_sm
from pycce.utilities import expand, outer, shorten_dimensions, gen_state_list


class RunObject:
    r"""
    Abstract class of the CCE simulation runner.

    Implements cluster correlation expansion, interlaced averaging, and sampling over random bath states.
    Requires definition of the following methods, from which the kernel will be automatically created:

        - ``.generate_hamiltonian(self)`` method which, using the attributes of the ``self`` object,
          computes cluster hamiltonian stored in ``self.cluster_hamiltonian``.

        - ``.compute_result(self)`` method which, using the attributes of the ``self``, computes the resulting
          quantity for the given cluster.

    Alternatively, user can define the kernel manually. Then the following methods have to be overridden:

        - ``.kernel(self, cluster, *args, **kwargs)`` method which takes indexes of the bath spins in the given
          cluster as a first positional argument. This method is required for usual CCE runs.

        - ``.interlaced_kernel(self, cluster, supercluster, *args, **kwargs)`` method which takes
          indexes of the bath spins in the given cluster as a first positional argument, indexes of the supercluster
          as a second positional argument. This method is required for interlaced CCE runs.

    Args:

        timespace (ndarray with shape (t, )): Time delay values at which to compute propagators.

        clusters (dict):
            Clusters included in different CCE orders of structure ``{int order: ndarray([[i,j],[i,j]])}``.

        bath (BathArray with shape (n,)): Array of *n* bath spins.

        magnetic_field (ndarray): Magnetic field of type ``magnetic_field = np.array([Bx, By, Bz])``.

        alpha (int or ndarray with shape (2s+1, )): :math:`\ket{0}` state of the qubit in :math:`S_z`
            basis or the index of eigenstate to be used as one.

        beta (int or ndarray with shape (2s+1, )): :math:`\ket{1}` state of the qubit in :math:`S_z` basis
            or the index of the eigenstate to be used as one.

        state (ndarray with shape (2s+1, )):
            Initial state of the central spin, used in gCCE and noise autocorrelation calculations.
            Defaults to :math:`\frac{1}{N}(\ket{0} + \ket{1})` if not set **OR** if alpha and beta are provided as
            indexes.


        spin (float): Value of the central spin.

        zfs (ndarray with shape (3,3)): Zero Field Splitting tensor of the central spin.

        gyro (float or ndarray with shape (3, 3)):
            Gyromagnetic ratio of the central spin

            **OR**

            tensor corresponding to interaction between magnetic field and
            central spin.

        nbstates (int): Number of random bath states to sample over in bath state sampling runs.

        bath_state (ndarray): Array of bath states in any accepted format.

        seed (int): Seed for the random number generator in bath states sampling.

        masked (bool):
            True if mask numerically unstable points (with result > result[0]) in the sampling over bath states
            False if not. Default True.

        projected_bath_state (ndarray with shape (n,)):
            Array with z-projections of the bath spins states.
            Overridden in runs with random bath state sampling.

        parallel (bool):
            True if parallelize calculation of cluster contributions over different mpi processes.
            Default False.

        direct (bool):
            True if use direct approach in run (requires way more memory but might be more numerically stable).
            False if use memory efficient approach. Default False.

        parallel_states (bool):
            True if use MPI to parallelize the calculations of density matrix
            for each random bath state.

        **kwargs: Additional keyword arguments to be set as the attributes of the given object.

    """

    #        fixstates (dict):
    #             If provided, contains indexes of bath spins with fixed pure state for sampling runs and interlaced runs.
    #
    #             Each key is the index of bath spin,
    #             value - fixed :math:`\hat{I}_z` projection of the **pure** :math:`\hat{I}_z` eigenstate of bath spin.

    result_operator = operator.imul
    """Operator which will combine the result of expansion,.
    
    Default: ``operator.imul``."""
    contribution_operator = operator.ipow
    """Operator which will combine multiple contributions of the same cluster 
    in the optimized approach.
    
    Default: ``operator.ipow``."""
    removal_operator = operator.itruediv
    """Operator which will remove subcluster contribution from the given cluster contribution.
    First argument cluster contribution, second - subcluster contribution.
    
    Defalut: ``operator.itruediv``."""
    addition_operator = np.prod
    """Group operation which will combine contributions from the different clusters into
    one contribution in the direct approach.
    
    Default: ``numpy.prod``."""

    def __init__(self, timespace,
                 clusters, bath,
                 magnetic_field,
                 center=None,
                 pulses=None,
                 nbstates=None,
                 seed=None,
                 masked=True,
                 parallel=False,
                 direct=False,
                 parallel_states=False,
                 **kwargs):

        self.nbstates = nbstates
        """int: Number of random bath states to sample over in bath state sampling runs."""
        self.timespace = timespace
        """ndarray with shape (t, ): Time points at which result will be computed."""
        self.clusters = clusters
        """dict: Clusters included in different CCE orders of structure ``{int order: ndarray([[i,j],[i,j]])}``."""
        self.bath = bath
        """BathArray with shape (n,): Array of *n* bath spins."""
        self.center = center
        """CenterArray: Properties of the central spin."""
        # self.projected_bath_state = projected_bath_state
        # """ndarray with shape (n,): Array with z-projections of the bath spins states.
        # Overridden in runs with random bath state sampling."""
        self.magnetic_field = magnetic_field
        """ndarray or callable: Magnetic field of type ``magnetic_field = np.array([Bx, By, Bz])``,
        or a function that takes position as an argument."""

        self.parallel = parallel
        """bool: True if parallelize calculation of cluster contributions over different mpi processes.
        Default False."""
        self.parallel_states = parallel_states
        """bool: True if use MPI to parallelize the calculations of density matrix
        for each random bath state."""
        self.direct = direct
        """bool: True if use direct approach in run (requires way more memory but might be more numerically stable).
        False if use memory efficient approach. Default False."""
        # MC Bath state sampling parameters
        self.seed = seed
        """int:  Seed for the random number generator in bath states sampling."""
        self.masked = masked
        """bool: True if mask numerically unstable points (with result > result[0]) in the sampling over bath states
        False if not. Default True."""

        for k in kwargs:
            setattr(self, k, kwargs[k])

        # self._cluster_hamiltonian = None

        self.hamiltonian = None
        """ndarray: Full cluster Hamiltonian."""

        self.cluster = None
        """BathArray: Array of the bath spins inside the given cluster."""
        self.cluster_indexes = None
        # self.others = None
        # """BathArray: Array of the bath spins outside the given cluster."""
        # self.others_mask = None
        # """ndarray: Bool array of of size self.bath with True entries for each of the spin outside of the cluster."""
        self.states = None

        self.has_states = False
        """bool: Whether there are states provided in the bath during the run."""
        self.initial_states_mask = bath.has_state
        """ndarray: Bool array of the states, initially present in the bath."""
        self.pulses = pulses
        """ Sequence: Sequence object, containing series of pulses, applied to the system."""

        self.projected_states = None
        """ndarray: Array of :math:`S_z` projections of the bath spins after each control pulse, 
        involving bath spins.
        """
        self.base_hamiltonian = None
        """Hamiltonian: Hamiltonian of the given cluster without mean field additions.
        In conventional CCE, also excludes additions from central spins."""
        self.result = None
        """ndarray: Result of the calculation."""
        self.delays = None
        """list or None: List with delays before each pulse or None if equispaced.
        Generated by ``.generate_pulses`` method."""
        self.rotations = None
        """list: List with matrix representations of the rotation from each pulse.
        Generated by ``.generate_pulses`` method."""

    def preprocess(self):
        """
        Method which will be called before cluster-expanded run.
        """

        self.base_hamiltonian = None
        self.hamiltonian = None

        self.cluster = None
        self.cluster_indexes = None
        # self.others = None
        # self.others_mask = None

        self.states = None
        self.projected_states = None

        self.result = None
        self.delays = None
        self.rotations = None

        self.has_states = self.bath.state.any()

        if self.has_states:
            bath = self.bath
            # proj = self.bath.proj
            if isinstance(self.pulses, Sequence) and any(p.bath_names is not None for p in self.pulses):
                self.projected_states = generate_rotated_projected_states(bath, self.pulses)

        else:
            bath = None
            # proj = None

        self.center.generate_states(self.magnetic_field, bath=bath)

    def postprocess(self):
        """
        Method which will be called after cluster-expanded run.
        """
        pass

    def generate_hamiltonian(self):
        raise NotImplementedError

    def compute_result(self):
        raise NotImplementedError

    def kernel(self, cluster, *args, **kwargs):
        """
        Central kernel that will be called in the cluster-expanded calculations.

        Args:
            cluster (ndarray): Indexes of the bath spins in the given cluster.
            *args: Positional arguments of the kernel.
            **kwargs: Keyword arguments of the kernel.

        Returns:

            ndarray: Results of the calculations.

        """
        self.cluster_indexes = cluster
        self.cluster = self.bath[cluster]

        if self.has_states:
            self.states = self.cluster.state
        else:
            self.states = None

        self.base_hamiltonian = self.generate_hamiltonian()
        self._check_hamiltonian()
        if isinstance(self.pulses, Sequence):
            self.generate_pulses()

        result = self.compute_result()

        return result

    @property
    def __inner_kernel(self):
        return cluster_expansion_decorator(self.kernel,
                                           result_operator=self.result_operator,
                                           contribution_operator=self.contribution_operator,
                                           removal_operator=self.removal_operator,
                                           addition_operator=self.addition_operator)

    def run(self, *args, **kwargs):
        """
        Method that runs cluster-expanded single calculation.

        Args:
            *args: Positional arguments of the kernel.
            **kwargs: Keyword arguments of the kernel.

        Returns:

            ndarray: Results of the calculations.

        """
        self.preprocess()
        self.result = self.__inner_kernel(self, *args, **kwargs)
        self.postprocess()
        return self.result

    @monte_carlo_method_decorator
    def __inner_sampled_run(self, *args, **kwargs):
        return self.run(*args, **kwargs)

    def sampling_run(self, *args, **kwargs):
        """
        Method that runs bath sampling calculations.

        Args:
            *args: Positional arguments of the kernel.
            **kwargs: Keyword arguments of the kernel.

        Returns:

            ndarray: Results of the calculations.

        """
        self.result = self.__inner_sampled_run(*args,
                                               **kwargs)
        return self.result

    def interlaced_kernel(self, cluster, supercluster, *args, **kwargs):
        """
        Central kernel that will be called in the cluster-expanded calculations
        with interlaced averaging of bath spin states.

        Args:
            cluster (ndarray): Indexes of the bath spins in the given cluster.

            supercluster (ndarray): Indexes of the bath spins in the supercluster of the given cluster.
                Supercluster is the union of all clusters in ``.clusters`` attribute, for which given cluster
                is a subset.

            *args: Positional arguments of the kernel.

            **kwargs: Keyword arguments of the kernel.

        Returns:

            ndarray: Results of the calculations.

        """
        sc_mask = ~np.isin(supercluster, cluster)

        self.cluster_indexes = cluster
        self.cluster = self.bath[cluster]

        if not sc_mask.any():
            result = self.kernel(cluster, *args, **kwargs)
            return result

        if self.has_states:
            self.states = self.cluster.state
        else:
            self.states = None

        self.base_hamiltonian = self.generate_hamiltonian()
        if isinstance(self.pulses, Sequence):
            self.generate_pulses()

        outer_indexes = supercluster[sc_mask]
        outer_spin = self.bath[outer_indexes]

        index_state = 0
        initial_outer_spin_state = outer_spin.state

        result = 0

        for index_state, state in enumerate(self.generate_supercluser_states(outer_indexes)):

            self.bath.state[outer_indexes] = gen_state_list(state, outer_spin.dim)
            # Note that if put others in different line the updated bath state might not copy properly

            if self.projected_states is not None:
                self.projected_states[outer_indexes] = generate_rotated_projected_states(self.bath[outer_indexes],
                                                                                         self.pulses)

            self._check_hamiltonian()
            result += self.compute_result()

        self.bath.state[outer_indexes] = initial_outer_spin_state

        result /= index_state + 1

        return result

    @property
    def __inner_interlaced_kernel(self):
        return interlaced_decorator(self.interlaced_kernel,
                                    result_operator=self.result_operator,
                                    contribution_operator=self.contribution_operator)

    def interlaced_run(self, *args, **kwargs):
        """
        Method that runs cluster-expanded single calculation with interlaced averaging of bath spin states.

        Args:
            *args: Positional arguments of the interlaced kernel.
            **kwargs: Keyword arguments of the interlaced kernel.

        Returns:

            ndarray: Results of the calculations.

        """
        self.preprocess()
        self.result = self.__inner_interlaced_kernel(self, *args, **kwargs)
        self.postprocess()

        return self.result

    @monte_carlo_method_decorator
    def __inner_sampled_interlaced_run(self, *args, **kwargs):
        return self.interlaced_run(*args, **kwargs)

    def sampling_interlaced_run(self, *args, **kwargs):
        """
        Method that runs bath sampling calculations with interlaced averaging of bath spin states.

        Args:
            *args: Positional arguments of the interlaced kernel.
            **kwargs: Keyword arguments of the interlaced kernel.

        Returns:

            ndarray: Results of the calculations.

        """
        self.result = self.__inner_sampled_interlaced_run(*args,
                                                          **kwargs)
        return self.result

    @classmethod
    def from_simulator(cls, sim, **kwargs):
        r"""
        Class method to generate ``RunObject`` from the properties of ``Simulator`` object.

        Args:
            sim (Simulator): Object, whose properties will be used to initialize ``RunObject`` instance.
            **kwargs: Additional keyword arguments that will replace ones, recovered from the ``Simulator`` object.

        Returns:
            RunObject: New instance of ``RunObject`` class.

        """
        parameters = vars(sim).copy()

        for k in list(parameters.keys()):
            if k[0] == '_':
                parameters[k[1:]] = parameters.pop(k)

        parameters.update(**kwargs)
        run = cls(**parameters)

        return run

    def generate_supercluser_states(self, supercluster):
        """
        Helper function to generate all possible pure states of the given supercluster.

        Args:
            supercluster (ndarray with shape (n, )): Indexes of the bath spins in the supercluster.

        Yields:
            ndarray with shape (n, ): Pure state of the given supercluster.

        """
        sc = self.bath[supercluster]
        his = self.initial_states_mask[supercluster]  # have initial states

        if not his.any():
            states = np.asarray(np.meshgrid(*[np.linspace(-s.s, s.s, s.dim) for s in sc])).T.reshape(-1, sc.size)
        else:
            # one liner I'm hecking proud of
            # 1) generate list of linspaces spanning the possible Sz projections if no initial state provided
            #    othervise give an zero-sized array of the projection
            # 2) Create meshgrid out of this list with the dimensions (sc.size, dim_1, dim_2, 1, dim_4, ..., dim_n)
            #    where dim is dimensions of the nuclear spin and 1 in case only one projection is given
            # 3) Transpose to change to (dim_n, dim_n-1, ..., dim_4, 1, dim_2, dim_1, sc.size)
            states = np.asarray(np.meshgrid(
                *[np.linspace(-sc[i].s, sc[i].s, sc[i].dim) if not his[i] else sc[i].proj for i in
                  range(sc.size)])).T.reshape(-1, sc.size)

        for single in states:
            yield single

    def generate_pulses(self):
        """
        Generate list of matrix representations of the rotations, induced by the sequence of the pulses.

        Returns:
            tuple: *tuple* containing:

            * **list** or **None**: List with delays before each pulse or None if equispaced.

            * **list**: List with matrix representations of the rotation from each pulse.
        """
        dimensions = self.base_hamiltonian.dimensions
        vectors = self.base_hamiltonian.vectors
        cs = self.center
        bath = self.cluster
        ndims = len(dimensions)
        bathvectors = vectors[:bath.shape[0]]

        shortdims = None
        csindex = None

        if cs is not None:
            if bath is not None and ndims == bath.size:
                cs = None  # Ignore central spin
            else:
                nc = len(cs)  # number of central spins
                shortdims = shorten_dimensions(dimensions, nc)
                csindex = ndims - nc

        self.delays = None
        self.rotations = None

        equispaced = not any(p._has_delay for p in self.pulses)

        if equispaced:
            delays = None

        else:
            delays = [p.delay if p.delay is not None else 0 for p in self.pulses]

        rots = []
        # Sigma as if central spin array is total spin. Sigma - pauli matrix

        total_svec = None
        separate_svec = dict()

        for p in self.pulses:

            initial_rotation = rotation = 1

            if p.naxes and cs is not None:
                if p.which is None:

                    if total_svec is None:
                        total_svec = from_sigma(cs.sigma, csindex, shortdims) / 2
                    rotation = p.generate_rotation(total_svec, spin_half=True)

                else:
                    for i in p.which:
                        c = cs[i]
                        if i not in separate_svec:
                            separate_svec[i] = from_sigma(c.sigma, csindex, shortdims) / 2
                        rotation = np.dot(p.generate_rotation(separate_svec[i], spin_half=True),
                                          rotation)

            for name in p:

                which = (bath.N == name)

                if any(which):
                    vecs = bathvectors[which]
                    rotation = np.dot(pulse_bath_rotation(p[name], vecs), rotation)

            if initial_rotation is rotation:
                rotation = None

            p.rotation = rotation

            rots.append(rotation)

        self.delays = delays
        self.rotations = rots

        return delays, rots

    def _check_hamiltonian(self):
        self.hamiltonian = None
        # if no projected states hamiltonian is simple
        if self.projected_states is None:

            if self.has_states:
                addition = external_spins_field(self.base_hamiltonian.vectors, self.cluster_indexes, self.bath,
                                                self.bath.proj)
                # addition = zero_order_addition(self.base_hamiltonian.vectors, self.cluster, self.others,
                #                                self.others.proj)
                self.hamiltonian = self.base_hamiltonian.data + addition

            else:
                self.hamiltonian = self.base_hamiltonian.data

    def get_hamiltonian_variable_bath_state(self, index=0):
        """
        Generate Hamiltonian in case of the complicated pulse sequence.

        Args:
            index (int): Index of the flips of spin states.

        Returns:
            ndarray with shape (n, n): Hamiltonian with mean field additions from the given set of projected states.
        """
        if self.projected_states is None:
            return self.hamiltonian

        if self.has_states:
            # addition = zero_order_addition(self.base_hamiltonian.vectors, self.cluster, self.others,
            #                                self.projected_states[self.others_mask, index])
            addition = external_spins_field(self.base_hamiltonian.vectors, self.cluster_indexes, self.bath,
                                            self.projected_states[:, index])
        else:
            addition = 0

        self.hamiltonian = self.base_hamiltonian.data + addition

        return self.hamiltonian


def from_sigma(sigma, i, dims):
    """
    Generate spin vector from dictionary with spin matrices.

    Args:
        sigma (dict): Dictionary, which contains spin matrices of form ``{'x': Sx, 'y': Sy, 'z': Sz}``.
        i (int): Index of the spin in the order of ``dims``.
        dims (ndarray with shape (N,)): Dimensions of the spins in the given cluster.

    Returns:
        ndarray with shape (3, n, n): Spin vector in a full Hilbert space.

    """
    return np.array([expand(sigma[ax], i, dims) for ax in sigma])


def generate_rotated_projected_states(bath, pulses):
    """
    Generate projected states after each control pulse, involving bath spins.

    Args:
        bath (BathArray with shape (n, )): Array of bath spins.
        pulses (Sequence): Sequence of pulses.

    Returns:
        ndarray with shape (n, x):
            Array of :math:`S_z` projections of bath spin states after each pulse, involving bath spins.
            Each :math:`i`-th column is projections before the :math:`i`-th pulse involving bath spins.

    """
    state = bath.state
    projected = [state.proj]
    rotations = {}
    for p in pulses:

        if p.bath_names is not None:
            proj = state.proj.copy()

            for name in p:

                which = (bath.N == name)

                if any(which):
                    spin = bath.types[name].s

                    rotation = p[name].generate_rotation(np.array(numba_gen_sm(round(2 * spin + 1))))

                    if name in rotations:
                        rotation = rotation @ rotations[name]

                    rotations[name] = rotation
                    proj[which] = bath[which].state.project(rotation)

            projected.append(proj)

    return np.swapaxes(projected, 0, -1)


_rot = {'x': 0, 'y': 1, 'z': 2}


def pulse_bath_rotation(pulse, vectors):
    """
    Generate rotation of the bath spins from the given pulse.

    Args:
        pulse (Pulse): Control pulse.
        vectors (ndarray with shape (n, 3, N, N): Array of spin vectors.

    Returns:
        ndarray with shape (x, x): Matrix representation of the spin rotation.

    """
    rotation = pulse.generate_rotation(vectors[0], spin_half=vectors[0, 0, 0, 0] < 1)

    for v in vectors[1:]:
        np.matmul(rotation, pulse.generate_rotation(v, spin_half=v[0, 0, 0] < 1), out=rotation)

    return rotation


def simple_propagator(timespace, hamiltonian):
    r"""
    Generate a simple propagator :math:`U=\exp[-\frac{i}{\hbar} \hat H]` from the Hamiltonian.

    Args:

        timespace (ndarray with shape (n, )): Time points at which to evaluate the propagator.
        hamiltonian (ndarray with shape (N, N)): Hamiltonian of the system.

    Returns:
        ndarray with shape (n, N, N): Propagators, evaluated at each timepoint.
    """
    evalues, evec = np.linalg.eigh(hamiltonian * PI2)

    eigexp = np.exp(-1j * np.outer(timespace, evalues),
                    dtype=np.complex128)

    return np.matmul(np.einsum('...ij,...j->...ij', evec, eigexp, dtype=np.complex128),
                     evec.conj().T)


@jit(cache=True, nopython=True)
def from_central_state(dimensions, central_state):
    """
    Generate density matrix of the system if all spins apart from central spin are in completely mixed state.

    Args:
        dimensions (ndarray with shape (n,)): Array of the dimensions of the spins in the cluster.
        central_state (ndarray with shape (x,)): Density matrix of central spins.

    Returns:
        ndarray with shape (N, N): Density matrix for the whole cluster.
    """

    return expand(central_state, len(dimensions) - 1, dimensions) / dimensions[:-1].prod()


@jit(cache=True, nopython=True)
def from_none(dimensions):
    """
    Generate density matrix of the systems if all spins are in completely mixed state.
    Args:
        dimensions (ndarray with shape (n,)): Array of the dimensions of the spins in the cluster.

    Returns:
        ndarray with shape (N, N): Density matrix for the whole cluster.

    """
    tdim = np.prod(dimensions)
    return np.eye(tdim) / tdim


def from_states(states):
    """
    Generate density matrix of the systems if all spins are in pure states.
    Args:
        states (array-like): Array of the pure spin states.

    Returns:
        ndarray with shape (N, N): Spin vector for the whole cluster.

    """
    cluster_state = states[0]
    for s in states[1:]:
        cluster_state = np.kron(cluster_state, s)

    return cluster_state


def combine_cluster_central(cluster_state, central_state):
    """
    Combine bath spin states and the state of central spin.
    Args:
        cluster_state (ndarray with shape (n,) or (n, n)): State vector or density matrix of the bath spins.
        central_state (ndarray with shape (m,) or (m, m)): State vector or density matrix of the central spins.

    Returns:
        ndarray with shape (mn, ) or (mn, mn): State vector or density matrix of the full system.
    """
    lcs = len(cluster_state.shape)
    ls = len(central_state.shape)

    if lcs != ls:
        return _noneq_cc(cluster_state, central_state)
    else:
        return _eq_cc(cluster_state, central_state)


@jit(cache=True, nopython=True)
def _noneq_cc(cluster_state, central_state):
    if len(cluster_state.shape) == 1:
        matrix = outer(cluster_state, cluster_state)
        return np.kron(matrix, central_state)

    else:
        matrix = outer(central_state, central_state)
        return np.kron(cluster_state, matrix)


@jit(cache=True, nopython=True)
def _eq_cc(cluster_state, central_state):
    return np.kron(cluster_state, central_state)


@jit(cache=True, nopython=True)
def rand_state(d):
    """
    Generate random state of the spin.

    Args:
        d (int): Dimensions of the spin.

    Returns:
        ndarray with shape (d, d): Density matrix of the random state.
    """
    return np.eye(d, dtype=np.complex128) / d


def generate_initial_state(dimensions, states=None, central_state=None):
    """
    Generate initial state of the cluster.

    Args:
        dimensions (ndarray with shape (n, )): Dimensions of all spins in the cluster.
        states (BathState, optional): States of the bath spins. If None, assumes completely random state.
        central_state (ndarray): State of the central spin. If None, assumes that no central spin is present
            in the Hilbert space of the cluster.

    Returns:
        ndarray with shape (N,) or (N, N): State vector or density matrix of the cluster.

    """
    if states is None:
        if central_state is None:
            return from_none(dimensions)
        else:
            if len(central_state.shape) == 1:
                central_state = outer(central_state, central_state)
            return from_central_state(dimensions, central_state)

    has_none = not states.has_state.all()
    all_pure = False
    all_mixed = False

    if not has_none:
        all_pure = states.pure.all()
        if not all_pure:
            all_mixed = (~states.pure).all()

    if has_none:
        for i in range(states.size):
            if states[i] is None:
                states[i] = rand_state(dimensions[i])

    if not (all_pure or all_mixed):
        for i in range(states.size):

            if len(states[i].shape) < 2:
                states[i] = outer(states[i], states[i])

    cluster_state = from_states(states)

    if central_state is not None:
        cluster_state = combine_cluster_central(cluster_state, central_state)

    return cluster_state



import numpy as np
from pycce.constants import PI2
from pycce.h.functions import projected_addition
from pycce.h.total import bath_hamiltonian
from pycce.run.base import RunObject, simple_propagator, generate_initial_state


def _rotmul(rotation, u, **kwargs):
    if rotation is not None and u is not None:
        u = np.matmul(rotation, u, **kwargs)
    elif rotation is not None:
        u = rotation

    return u


def simple_propagators(delays, hamiltonian_alpha, hamiltonian_beta):
    r"""
    Generate two simple propagators :math:`U=\exp[-\frac{i}{\hbar} \hat H]` from the Hamiltonians, conditioned
    on two qubit levels.

    Args:
        delays (ndarray with shape (n, )): Time points at which to evaluate the propagator.
        hamiltonian_alpha (ndarray with shape (N, N)): Hamiltonian of the bath spins with qubit in alpha state.
        hamiltonian_beta (ndarray with shape (N, N)): Hamiltonian of the bath spins with qubit in beta state.

    Returns:
        tuple:
            * **ndarray with shape (n, N, N)**:
              Matrix representation of the propagator conditioned on the alpha qubit state for each time point.
            * **ndarray with shape (n, N, N)**:
              Matrix representation of the propagator conditioned on the beta qubit state for each time point.

    """
    u0 = simple_propagator(delays, hamiltonian_alpha)
    u1 = simple_propagator(delays, hamiltonian_beta)

    return u0, u1


def propagate_propagators(v0, v1, number):
    """
    From two simple propagators and number of pulses in CPMG sequence generate two full propagators.
    Args:
        v0 (ndarray with shape (n, N, N)): Propagator conditioned on the alpha qubit state for each time point.
        v1 (ndarray with shape (n, N, N)): Propagator conditioned on the beta qubit state for each time point.
        number (int): Number of pulses.

    Returns:
        tuple:
            * **ndarray with shape (n, N, N)**:
              Matrix representation of the propagator conditioned on the alpha qubit state for each time point.
            * **ndarray with shape (n, N, N)**:
              Matrix representation of the propagator conditioned on the beta qubit state for each time point.

    """
    v0_he = np.matmul(v0, v1, dtype=np.complex128)
    v1_he = np.matmul(v1, v0, dtype=np.complex128)

    if number == 1:
        return v0_he, v1_he

    v0_cp = np.matmul(v0_he, v1_he, dtype=np.complex128)  # v0 @ v1 @ v1 @ v0
    v1_cp = np.matmul(v1_he, v0_he, dtype=np.complex128)  # v1 @ v0 @ v0 @ v1

    unitary_0 = np.linalg.matrix_power(v0_cp, number // 2)
    unitary_1 = np.linalg.matrix_power(v1_cp, number // 2)

    if number % 2 == 1:
        unitary_0 = np.matmul(unitary_0, v0_he)
        unitary_1 = np.matmul(unitary_1, v1_he)

    return unitary_0, unitary_1


def _close_state_index(state, eiv, level_confidence=0.95):
    """
    Get index of the eigenstate stored in eiv,
    which has fidelity higher than ``level_confidence`` with the provided ``state``.

    Args:
        state (ndarray with shape (2s+1,)): State for which to find the analogous eigen state.
        eiv (ndarray with shape (2s+1, 2s+1)): Matrix of eigenvectors as columns.
        level_confidence (float): Threshold fidelity. Default 0.95.

    Returns:
        int: Index of the eigenstate.
    """
    indexes = np.argwhere((eiv.T @ state) ** 2 > level_confidence).flatten()

    if not indexes.size:
        raise ValueError(f"Initial qubit state is below F = {level_confidence} "
                         f"to the eigenstate of central spin Hamiltonian.\n"
                         f"Qubit level:\n{repr(state)}"
                         f"Eigenstates (rows):\n{repr(eiv.T)}")
    return indexes[0]


class CCE(RunObject):
    """
    Class for running conventional CCE simulations.

    .. note::

        Subclass of the ``RunObject`` abstract class.

    Args:
        *args: Positional arguments of the ``RunObject``.

        pulses (int or Sequence):
            number of pulses in CPMG sequence or instance of Sequence object.
            For now, only CPMG sequences are supported in conventional CCE simulations.

        as_delay (bool):
            True if time points are delay between pulses, False if time points are total time.

        second_order (bool):
            True if add second order perturbation theory correction to the cluster Hamiltonian.
            If set to True sets the qubit states as eigenstates of central spin Hamiltonian from the following
            procedure. If qubit states are provided as vectors in :math:`S_z` basis,
            for each qubit state compute the fidelity of the qubit state and
            all eigenstates of the central spin and chose the one with fidelity higher than ``level_confidence``.
            If such state is not found, raises an error.

        level_confidence (float): Maximum fidelity of the qubit state to be considered eigenstate of the
            central spin Hamiltonian. Default 0.95.

        **kwargs: Keyword arguments of the ``RunObject``.

    """

    def __init__(self, *args, as_delay=False, second_order=False,
                 level_confidence=0.95, **kwargs):
        self.key_alpha = None
        self.key_beta = None

        self.initial_pulses = None
        """int or Sequence: Input pulses"""
        self.pulses = None
        """int or Sequence: If input Sequence contains only pi pulses at even delay, stores number of pulses.
        Otherwise stores full ``Sequence``."""
        self.as_delay = as_delay
        """bool: True if time points are delay between pulses, False if time points are total time."""
        self.second_order = second_order
        """bool: True if add second order perturbation theory correction to the cluster hamiltonian."""
        self.level_confidence = level_confidence
        """float: Maximum fidelity of the qubit state to be considered eigenstate of the central spin hamiltonian."""
        self.energy_alpha = None
        """float: Eigen energy of the alpha state in the central spin Hamiltonian."""
        self.energy_beta = None
        """float: Eigen energy of the beta state in the central spin Hamiltonian."""
        self.energies = None
        """ndarray with shape (2s+1, ): All eigen energies of the central spin Hamiltonian."""
        self.projections_alpha_all = None
        r"""ndarray with shape (2s+1, 3): Array of vectors with spin operator matrix elements
        of type :math:`[\bra{0}\hat S_x\ket{i}, \bra{0}\hat S_y\ket{i}, \bra{0}\hat S_z\ket{i}]`, where
        :math:`\ket{0}` is the alpha qubit state,
        :math:`\ket{i}` are all eigenstates of the central spin hamiltonian."""
        self.projections_beta_all = None
        r"""ndarray with shape (2s+1, 3): Array of vectors with spin operator matrix elements
        of type :math:`[\bra{1}\hat S_x\ket{i}, \bra{1}\hat S_y\ket{i}, \bra{1}\hat S_z\ket{i}]`, where
        :math:`\ket{1}` is the beta qubit state,
        :math:`\ket{i}` are all eigenstates of the central spin hamiltonian."""
        self.projections_alpha = None
        r"""ndarray with shape (3,): Vector with spin operator matrix elements
        of type :math:`[\bra{0}\hat S_x\ket{0}, \bra{0}\hat S_y\ket{0}, \bra{0}\hat S_z\ket{0}]`, where
        :math:`\ket{0}` is the alpha qubit state"""
        self.projections_beta = None
        r"""ndarray with shape (3,): Vectors with spin operator matrix elements
        of type :math:`[\bra{1}\hat S_x\ket{1}, \bra{1}\hat S_y\ket{1}, \bra{1}\hat S_z\ket{1}]`, where
        :math:`\ket{1}` is the beta qubit state."""

        self.use_pulses = False
        """bool: True if use full ``Sequence``. False if use only number of pulses."""

        super().__init__(*args, **kwargs)

    def preprocess(self):
        super().preprocess()
        self.initial_pulses = self.pulses
        """ Sequence: Sequence object, containing series of pulses, applied to the system."""
        try:
            pulses = self.initial_pulses
            number = len(self.initial_pulses)

            angles = np.array([p._angles for p in pulses])
            naxes = [p.naxes for p in pulses]

            if not number:
                self.pulses = number

            else:
                c0 = all(p.which is None for p in pulses)
                c1 = all(n <= 1 for n in naxes)
                c2 = all([not p.keys() for p in pulses])
                c3 = all([not p._has_delay for p in pulses])
                c4 = np.isclose(angles, np.pi).any(axis=1).all()

                if c0 & c1 & c2 & c3 & c4:
                    self.pulses = number
                else:
                    self.pulses = self.initial_pulses
                    self.use_pulses = True

                if not c1 or not (np.isclose(angles, np.pi) | np.isclose(angles, 0)).all():
                    raise ValueError('Only pi-pulses are supported in CCE. Use gCCE for user-defined sequences.')

                if not c0 and self.second_order:
                    raise ValueError('Only full flip pulses are supported in CCE with second order corrections')

        except TypeError:
            self.pulses = self.initial_pulses

        self.center.generate_projections(second_order=self.second_order, level_confidence=self.level_confidence)

        if self.center.state_index is not None and isinstance(self.center.state_index, np.ndarray):
            self.key_alpha = self.center.state_index[:, 0]
            self.key_beta = self.center.state_index[:, 1]
        else:
            self.key_alpha = np.ones(len(self.center), dtype=bool)
            self.key_beta = np.zeros(len(self.center), dtype=bool)

    def postprocess(self):
        super().postprocess()
        self.h = None
        pass

    def generate_hamiltonian(self):
        """
        Using the attributes of the ``self`` object,
        compute the two projected cluster hamiltonians.

        Returns:

            tuple: Tuple containing:

                * **Hamiltonian**:
                  Cluster hamiltonian when qubit in the alpha state.
                * **Hamiltonian**:
                  Cluster hamiltonian when qubit in the alpha state.

        """
        hamil = bath_hamiltonian(self.cluster, self.magnetic_field)

        return hamil

    def compute_result(self):
        """
        Using the attributes of the ``self`` object,
        compute the coherence function as overlap in the bath evolution.

        Returns:

            ndarray: Computed coherence.

        """
        unitary_0, unitary_1 = self.propagators()

        if self.states is None:
            coherence_function = np.einsum('zij,zij->z', unitary_0, unitary_1.conj()) / unitary_0.shape[1]

        else:
            dm = generate_initial_state(self.base_hamiltonian.dimensions, states=self.states)
            # tripple einsum is slow
            # coherence_function = np.einsum('zli,ij,zlj->z', U0, dm, U1.conj())
            if len(dm.shape) > 1:

                dm_udagger = np.matmul(dm, unitary_1.conj().transpose(0, 2, 1))
                coherence_function = np.trace(np.matmul(unitary_0, dm_udagger), axis1=1, axis2=2)

            else:

                rightside = unitary_1 @ dm
                leftside = unitary_0 @ dm

                coherence_function = np.einsum('ki,ki->k', leftside.conj(), rightside)

        return coherence_function

    def propagators(self):
        """
        Generate two propagators, conditioned on the qubit state.

        Returns:
            tuple: *tuple* containing:

                * **ndarray with shape (t, n, n)**:
                  Matrix representation of the propagator conditioned on the alpha qubit state for each time point.
                * **ndarray with shape (t, n, n)**:
                  Matrix representation of the propagator conditioned on the beta qubit state for each time point.

        """
        if not self.use_pulses:
            return self._no_pulses()

        if self.delays is None:
            return self._no_delays()

        return self._delays()

    def _proj_ham(self, index=0, alpha=True, beta=False):
        self.get_hamiltonian_variable_bath_state(index)

        ha = self.hamiltonian + projected_addition(self.base_hamiltonian.vectors,
                                                   self.cluster, self.center, alpha)

        hb = self.hamiltonian + projected_addition(self.base_hamiltonian.vectors,
                                                   self.cluster, self.center, beta)

        return ha, hb

    def _no_pulses(self):
        ha, hb = self._proj_ham()
        delays = self.timespace / (2 * self.pulses) if ((not self.as_delay) and self.pulses) else self.timespace
        v0, v1 = simple_propagators(delays, ha, hb)

        if not self.pulses:
            return v0, v1

        else:
            return propagate_propagators(v0, v1, self.pulses)

    def _no_delays(self):
        delays = self.timespace if self.as_delay else self.timespace / (2 * len(self.pulses))

        key_alpha = list(self.key_alpha)
        key_beta = list(self.key_beta)

        ha, hb = self._proj_ham(alpha=key_alpha, beta=key_beta)

        v0, v1 = simple_propagators(delays, ha, hb)
        vs = {}

        vs[tuple(key_alpha)] = v0
        vs[tuple(key_beta)] = v1

        unitary_0 = np.eye(v0.shape[1], dtype=np.complex128)
        unitary_1 = np.eye(v1.shape[1], dtype=np.complex128)

        ps_counter = 0

        for p, rotation in zip(self.pulses, self.rotations):
            unitary_0 = np.matmul(v0, unitary_0)
            unitary_1 = np.matmul(v1, unitary_1)

            unitary_0 = _rotmul(rotation, unitary_0)
            unitary_1 = _rotmul(rotation, unitary_1)

            if p.bath_names is not None:
                ps_counter += 1
                vs.clear()

            key_alpha, key_beta = _gen_key(p, key_alpha, key_beta)

            try:
                v0, v1 = vs[tuple(key_alpha)], vs[tuple(key_beta)]

            except KeyError:

                ha, hb = self._proj_ham(index=ps_counter, alpha=key_alpha, beta=key_beta)
                v0, v1 = simple_propagators(delays, ha, hb)

                vs[tuple(key_alpha)] = v0
                vs[tuple(key_beta)] = v1

            unitary_0 = np.matmul(v0, unitary_0)
            unitary_1 = np.matmul(v1, unitary_1)

        return unitary_0, unitary_1

    def _delays(self):

        times = 0
        key_alpha = list(self.key_alpha)
        key_beta = list(self.key_beta)

        ha, hb = self._proj_ham(alpha=key_alpha, beta=key_beta)

        eval0, evec0 = np.linalg.eigh(ha * PI2)
        eval1, evec1 = np.linalg.eigh(hb * PI2)

        # for timesteps, rotation in zip(pulses.delays, pulses.rotations):
        eval_evec = {}
        eval_evec[tuple(key_alpha)] = eval0, evec0
        eval_evec[tuple(key_beta)] = eval1, evec1

        unitary_0 = None
        unitary_1 = None
        ps_counter = 0
        for p, delay, rotation in zip(self.pulses, self.delays, self.rotations):
            if np.any(delay):
                eigen_exp0 = np.exp(-1j * np.outer(delay, eval0), dtype=np.complex128)

                eigen_exp1 = np.exp(-1j * np.outer(delay, eval1), dtype=np.complex128)

                u0 = np.matmul(np.einsum('...ij,...j->...ij', evec0, eigen_exp0, dtype=np.complex128), evec0.conj().T)

                u1 = np.matmul(np.einsum('...ij,...j->...ij', evec1, eigen_exp1, dtype=np.complex128), evec1.conj().T)

                times += delay

                unitary_0 = _rotmul(rotation, u0) if unitary_0 is None else np.matmul(u0, _rotmul(rotation, unitary_0))
                unitary_1 = _rotmul(rotation, u1) if unitary_1 is None else np.matmul(u1, _rotmul(rotation, unitary_1))
            else:
                unitary_0 = _rotmul(rotation, unitary_0)
                unitary_1 = _rotmul(rotation, unitary_1)

            if p.bath_names is not None:
                ps_counter += 1
                eval_evec.clear()

            key_alpha, key_beta = _gen_key(p, key_alpha, key_beta)

            try:
                eval0, evec0 = eval_evec[tuple(key_alpha)]
                eval1, evec1 = eval_evec[tuple(key_beta)]
            except KeyError:
                ha, hb = self._proj_ham(index=ps_counter, alpha=key_alpha, beta=key_beta)
                eval0, evec0 = np.linalg.eigh(ha * PI2)
                eval1, evec1 = np.linalg.eigh(hb * PI2)

                eval_evec[tuple(key_alpha)] = eval0, evec0
                eval_evec[tuple(key_beta)] = eval1, evec1

        which = np.isclose(self.timespace, times)
        if ((self.timespace - times)[~which] >= 0).all():

            eigen_exp0 = np.exp(-1j * np.outer(self.timespace - times,
                                               eval0), dtype=np.complex128)
            eigen_exp1 = np.exp(-1j * np.outer(self.timespace - times,
                                               eval1), dtype=np.complex128)

            u0 = np.matmul(np.einsum('ij,kj->kij', evec0, eigen_exp0, dtype=np.complex128),
                           evec0.conj().T)

            u1 = np.matmul(np.einsum('ij,kj->kij', evec1, eigen_exp1, dtype=np.complex128),
                           evec1.conj().T)

            unitary_0 = np.matmul(u0, unitary_0)
            unitary_1 = np.matmul(u1, unitary_1)

        elif not which.all():
            raise ValueError(f"Pulse sequence time steps add up to larger than total times"
                             f"{np.argwhere((self.timespace - times) < 0)} are longer than total time.")

        return unitary_0, unitary_1


def _gen_key(p, key_alpha, key_beta):
    if p.flip:
        if p.which is None:
            key_alpha, key_beta = key_beta, key_alpha
        else:
            for index in p.which:
                key_alpha[index], key_beta[index] = key_beta[index], key_alpha[index]
    return key_alpha, key_beta

# def compute_coherence(H0, H1, timespace, N, as_delay=False, states=None):
#     """
#     Function to compute cluster coherence function in conventional CCE.
#
#     Args:
#         H0 (ndarray): Hamiltonian projected on alpha qubit state.
#         H1 (ndarray): Hamiltonian projected on beta qubit state.
#         timespace (ndarray): Time points at which to compute coherence function.
#         N (int): Number of pulses in CPMG.
#         as_delay (bool):
#             True if time points are delay between pulses,
#             False if time points are total time.
#         states (ndarray): ndarray of bath states in any accepted format.
#
#     Returns:
#         ndarray: Coherence function of the central spin.
#
#     """
#     # if timespace was given not as delay between pulses,
#     # divide to obtain the delay
#     coherence_function = None
#     return coherence_function

# def propagators(timespace, H0, H1, pulses, as_delay=False):
#     """
#     Function to compute propagators U0 and U1 in conventional CCE.
#
#     Args:
#         timespace (ndarray with shape (t, )): Time delay values at which to compute propagators.
#
#         H0 (ndarray with shape (n, n)): Hamiltonian projected on alpha qubit state.
#
#         H1 (ndarray with shape (n, n)): Hamiltonian projected on beta qubit state.
#
#         pulses (int or Sequence): Sequence of pulses.
#
#         as_delay (bool):
#             True if time points are delay between pulses.
#             False if time points are total time.
#
#     Returns:
#         tuple: *tuple* containing:
#
#             * **ndarray with shape (t, n, n)**:
#               Matrix representation of the propagator conditioned on the alpha qubit state for each time point.
#             * **ndarray with shape (t, n, n)**:
#               Matrix representation of the propagator conditioned on the beta qubit state for each time point.
#
#     """
#     try:
#         number = len(pulses)
#         check = (not as_delay) & (pulses.delays is None)
#         use_rotations = True
#     except TypeError:
#         number = pulses
#         check = (not as_delay)
#         use_rotations = False
#
#     if check and number:
#         timespace = timespace / (2 * number)
#
#     if not use_rotations:
#         v0, v1 = simple_propagators(timespace, H0, H1)
#
#         if not number:
#             return v0, v1
#         else:
#             return propagate_propagators(v0, v1, number)
#
#     if pulses.delays is None:
#
#         v0, v1 = simple_propagators(timespace, H0, H1)
#
#         vs = [v0, v1]
#
#         U0 = np.eye(v0.shape[1], dtype=np.complex128)
#         U1 = np.eye(v1.shape[1], dtype=np.complex128)
#
#         Us = [U0, U1]
#
#         for p in pulses:
#             for i in range(2):
#                 Us[i] = np.matmul(vs[i], Us[i])
#                 Us[i] = _rotmul(p.rotation, Us[i])
#
#             if p.flip:
#                 vs = vs[::-1]
#
#             for i in range(2):
#                 Us[i] = np.matmul(vs[i], Us[i])
#
#         return Us[0], Us[1]
#
#     U0 = None
#     U1 = None
#
#     times = 0
#
#     eval0, evec0 = np.linalg.eigh(H0 * PI2)
#     eval1, evec1 = np.linalg.eigh(H1 * PI2)
#
#     # for timesteps, rotation in zip(pulses.delays, pulses.rotations):
#     evalues = [eval0, eval1]
#     evec = [evec0, evec1]
#
#     for p in pulses:
#         timesteps = p.delay
#         rotation = p.rotation
#         eigen_exp0 = np.exp(-1j * np.outer(timesteps,
#                                            evalues[0]), dtype=np.complex128)
#         eigen_exp1 = np.exp(-1j * np.outer(timesteps,
#                                            evalues[1]), dtype=np.complex128)
#
#         u0 = np.matmul(np.einsum('...ij,...j->...ij', evec[0], eigen_exp0, dtype=np.complex128),
#                        evec[0].conj().T)
#
#         u1 = np.matmul(np.einsum('...ij,...j->...ij', evec[1], eigen_exp1, dtype=np.complex128),
#                        evec[1].conj().T)
#
#         times += timesteps
#
#         if U0 is None:
#             U0 = _rotmul(rotation, u0)
#             U1 = _rotmul(rotation, u1)
#
#         else:
#             U0 = np.matmul(u0, U0)
#             U0 = _rotmul(rotation, U0)
#
#             U1 = np.matmul(u1, U1)
#             U1 = _rotmul(rotation, U1)
#
#         if p.flip:
#             evalues = evalues[::-1]
#             evec = evec[::-1]
#
#     which = np.isclose(timespace, times)
#     if ((timespace - times)[~which] >= 0).all():
#
#         eigen_exp0 = np.exp(-1j * np.outer(timespace - times,
#                                            evalues[0]), dtype=np.complex128)
#         eigen_exp1 = np.exp(-1j * np.outer(timespace - times,
#                                            evalues[1]), dtype=np.complex128)
#
#         u0 = np.matmul(np.einsum('ij,kj->kij', evec[0], eigen_exp0, dtype=np.complex128),
#                        evec[0].conj().T)
#
#         u1 = np.matmul(np.einsum('ij,kj->kij', evec[1], eigen_exp1, dtype=np.complex128),
#                        evec[1].conj().T)
#
#         U0 = np.matmul(u0, U0)
#         U1 = np.matmul(u1, U1)
#
#     elif not which.all():
#         raise ValueError(f"Pulse sequence time steps add up to larger than total times"
#                          f"{np.argwhere((timespace - times) < 0)} are longer than total time.")
#
#     return U0, U1
#



import numpy as np
from pycce.sm import dimensions_spinvectors, vecs_from_dims


class Hamiltonian:
    r"""
    Class containing properties of the Hamiltonian.

    Essentially wrapper for ndarray with additional attributes of ``dimensions`` and ``spins``.

    Usual methods (e.g. ``__setitem__`` or ``__getitem__``) access the ``data`` attribute.

    .. note::

        Algebraic operations with Hamiltonian will return ndarray instance.

    Args:
        dimensions (array-like): array of the dimensions for each spin in the Hilbert space of the Hamiltonian.

    Attributes:
        dimensions (ndarray): array of the dimensions for each spin in the Hilbert space of the Hamiltonian.
        spins (ndarray): array of the spins, spanning the Hilbert space of the Hamiltonian.
        vectors (list): list with spin vectors of form ``[[Ix, Iy, Iz], [Ix, Iy, Iz], ...]``.
        data (ndarray): matrix representation of the Hamiltonian.

    """

    def __init__(self, dimensions, vectors=None, data=None):
        self.dimensions = np.asarray(dimensions)
        self.spins = (dimensions - 1) / 2
        if vectors is None:
            vectors = vecs_from_dims(dimensions)
        self.vectors = vectors

        tdim = self.dimensions.prod()
        if data is None:
            self.data = np.zeros((tdim, tdim), dtype=np.complex128)
        else:
            assert data.shape == (tdim, tdim), "Wrong data shape"
            self.data = data.astype(np.complex128)

    def __getitem__(self, item):
        return self.data.__getitem__(item)

    def __setitem__(self, key, value):
        self.data.__setitem__(key, value)

    def __delitem__(self, key):
        self.data.__delitem__(key)

    def __getattr__(self, item):
        if item in dir(self):
            return getattr(self, item)
        else:
            return getattr(self.data, item)

    @classmethod
    def from_bath(cls, bath, center=None):
        dim, vectors = dimensions_spinvectors(bath, central_spin=center)
        return cls(dim, vectors=vectors)

    def __add__(self, other):
        res = self.data.__add__(other)
        return Hamiltonian(self.dimensions, self.vectors, res)

    def __iadd__(self, other):
        self.data.__iadd__(other)
        return self

    def __sub__(self, other):
        res = self.data.__sub__(other)
        return Hamiltonian(self.dimensions, self.vectors, res)

    def __isub__(self, other):
        self.data.__isub__(other)
        return self

    def __mul__(self, other):
        res = self.data.__mul__(other)
        return Hamiltonian(self.dimensions, self.vectors, res)

    def __imul__(self, other):
        self.data.__imul__(other)
        return self

    def __matmul__(self, other):
        res = self.data.__matmul__(other)
        return Hamiltonian(self.dimensions, self.vectors, res)

    def __imatmul__(self, other):
        self.data.__imatmul__(other)
        return self

    def __truediv__(self, other):
        res = self.data.__truediv__(other)
        return Hamiltonian(self.dimensions, self.vectors, res)

    def __itruediv__(self, other):
        self.data.__itruediv__(other)
        return self

    def __floordiv__(self, other):
        res = self.data.__floordiv__(other)
        return Hamiltonian(self.dimensions, self.vectors, res)

    def __ifloordiv__(self, other):
        self.data.__ifloordiv__(other)
        return self

    def __mod__(self, other):
        res = self.data.__mod__(other)
        return Hamiltonian(self.dimensions, self.vectors, res)

    def __imod__(self, other):
        self.data.__imod__(other)
        return self

    def __pow__(self, other):
        res = self.data.__pow__(other)
        return Hamiltonian(self.dimensions, self.vectors, res)

    def __ipow__(self, other):
        self.data.__ipow__(other)
        return self



import copy
import warnings

import numpy as np
from pycce.bath.array import common_isotopes, common_concentrations
from pycce.constants import BOHR_TO_ANGSTROM

coord_types = ['crystal', 'bohr', 'angstrom', 'alat']


class DFTCoordinates:
    r"""
    Abstract class of a container of the DFT output coordinates.

    Attributes:
        alat (float): The lattice parameter in angstrom.
        cell (ndarray with shape (3, 3)):
            cell is 3x3 matrix with entries:

            .. math::

                [&[a_x\ b_x\ c_x]\\
                &[a_y\ b_y\ c_y]\\
                &[a_z\ b_z\ c_z]]


            where a, b, c are crystallographic vectors
            and x, y, z are their coordinates in the cartesian reference frame.

        coordinates (ndarray with shape (n, 3)): Array with the coordinates of atoms in the cell.

        names (ndarray with shape (n,)): Array with the names of atoms in the cell.

        cell_units (str): Units of cell coordinates: 'bohr', 'angstrom', 'alat'.

        coordinates_units (str): Units of atom coordinates: 'crystal', 'bohr', 'angstrom', 'alat'.
    """

    def __init__(self):
        self.alat = None
        self.cell = None

        self.coordinates = None
        self.names = None
        self.cell_units = None
        self.coordinates_units = None

    def to_angstrom(self, inplace=False):
        """
        Method to transform cell and coordinates units to angstroms.

        Args:
            inplace (bool): if True changes attributes inplace. Otherwise returns copy.

        Returns:
            DFTCoordinates or subclass:
                Instance of the subclass with units of coordinates and cell of Angstroms.
        """
        if inplace:
            obj = self
        else:
            obj = copy.deepcopy(self)

        obj.cell = change_to_angstrom(self.cell.T, self.cell_units, alat=self.alat).T
        obj.coordinates = change_to_angstrom(self.coordinates, self.coordinates_units, alat=obj.alat, cell=obj.cell)
        obj.cell_units = 'angstrom'
        obj.coordinates_units = 'angstrom'
        return obj

    def get_angstrom(self, coordinate, units):
        """
        Change given coordinates to angstrom.

        Args:
            coordinates (ndarray with shape (n, 3) or (3,)): Coordinates to change.
            units (str): Initial units of the coordinates.

        Returns:
            ndarray (n, 3): Coordinates in angstrom.
        """
        return change_to_angstrom(coordinate, units, alat=self.alat, cell=self.cell)

    def __repr__(self):
        m = f"{type(self).__name__} with coordinates in {self.coordinates_units} and cell in {self.cell_units}."
        return m


def change_to_angstrom(coordinates, units, alat=None, cell=None):
    r"""
    Change coordinates to angstrom.

    Args:
        coordinates (ndarray with shape (n, 3) or (3,)): Coordinates to change.
        units (str): Initial units of the coordinates.
        alat (float): The lattice parameter in angstrom.

        cell (ndarray with shape (3,3)):
            cell is 3x3 matrix with entries:

            .. math::

                [&[a_x\ b_x\ c_x]\\
                &[a_y\ b_y\ c_y]\\
                &[a_z\ b_z\ c_z]]

            where a, b, c are crystallographic vectors,
            and x, y, z are their coordinates in the cartesian reference frame.

    Returns:
        ndarray with shape (n, 3): Coordinates in angstrom.

    """
    coordinates = np.asarray(coordinates)

    if units == 'angstrom':
        coordinates = coordinates
    if units == 'bohr':
        coordinates = coordinates * BOHR_TO_ANGSTROM
    if units == 'alat':
        coordinates = coordinates * alat
    if units == 'crystal':
        if len(coordinates.shape) > 1:
            coordinates = np.einsum('jk,ik->ij', cell, coordinates)
        else:
            coordinates = cell @ coordinates

    return coordinates


def fortran_value(value):
    """
    Get value from Fortran-type variable.

    Args:
        value (str): Value read from Fortran-type input.

    Returns:
        value (bool, str, float): value in Python format.

    """
    bools = {".true.": True, ".false": False}
    str_separators = "'" '"'

    if value in bools:
        value = bools[value]

    elif value.strip('+-').isdigit():
        value = int(value)

    elif value[0] in str_separators:
        value = value.strip(str_separators)

    else:
        try:
            value = float(value.replace('d', 'e'))
        except ValueError:
            raise ValueError(f'{value} is incorrect')

    return value


def yield_index(word, lines, start=0, case_sensitive=False):
    """
    Generator which yields indexes of the lines containing specific word.

    Args:
        word (str): Word to find in the line.
        lines (list of str): List of strings contraining lines from the file. Output of open(file).readlines().
        start (int): First index from which to start search.
        case_sensitive (bool): If True looks for the exact match. Otherwise the search is case insensitive.

    Yields:
        i (int): Index of the line containing word.
    """

    if not case_sensitive:
        word = word.lower()

    for i in range(start, len(lines)):
        lin = lines[i]
        if not case_sensitive:
            lin = lin.lower()

        if word in lin:
            yield i


def find_first_index(word, lines, start=0, case_sensitive=False):
    """
    Function to find first appearance of the index in the list of lines.

    Args:
        word (str): Word to find in the line.
        lines (list of str): List of strings contraining lines from the file. Output of open(file).readlines().
        start (int): First index from which to start search.
        case_sensitive (bool): If True looks for the exact match. Otherwise the search is case insensitive.

    Returns:
        i (int): Index of the first line from the start containing word.

    """
    try:
        ind = next(yield_index(word, lines, start=start, case_sensitive=case_sensitive))
    except StopIteration:
        raise KeyError(f'Could not find {word}')
    return ind


def set_isotopes(array, isotopes=None, inplace=True, spin_types=None):
    """
    Function to set the most common isotopes for the array containing DFT output. If some other isotope
    is specified, the A tensors are scaled accordingly.

    Args:
        array (BathArray): Array with DFT spins.
        isotopes (dict): Dictionary with chosen isotopes.
        inplace (bool): True if change the array inplace.
        spin_types (SpinDict): If provided, allows for custom defined ``SpinType`` instances.

    Returns:
        array (BathArray): Array with DFT spins with correct isotopes.

    """
    if not inplace:
        array = array.copy()

    if spin_types is None:
        spin_types = common_isotopes
    else:
        array.add_type(spin_types)

    for n in np.unique(array.N):
        try:
            isotopes_of_n = common_concentrations[n]
            names = list(isotopes_of_n.keys())
            conc = [isotopes_of_n[k] for k in names]
            which = names[np.argmax(conc)]

            if isotopes is not None and n in isotopes:
                factor = spin_types[isotopes[n]].gyro / spin_types[which].gyro
                which = isotopes[n]
            else:
                factor = 1

            where = array.N == n

            array.A[where] *= factor
            array.N[where] = which

            array.types[which]

        except KeyError:
            warnings.warn(f'Unknown element {n}. Skipping')
    return array



from collections.abc import MutableMapping
from itertools import combinations

import numpy as np
import scipy.sparse
import scipy.sparse.csgraph
from scipy.sparse import csr_matrix


class _Clusters(MutableMapping):
    """
    NOT IMPLEMENTED YET. Specific Class for storing the clusters objects
    """

    def __init__(self, ):
        self._data = {}

    def __getitem__(self, key):
        return self._data[key]

    def __delitem__(self, key):
        del self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return f"{type(self).__name__}(" + " ".join(str(x) for x in self.keys()) + ")"

    def keys(self):
        return self._data.keys()


def generate_clusters(bath, r_dipole, order, r_inner=0, ignore=None, strong=False, nclusters=None):
    """
    Generate clusters for the bath spins.

    Args:

        bath (BathArray): Array of bath spins.
        r_dipole (float): Maximum connectivity distance.
        order (int): Maximum size of the clusters to find.

        r_inner (float): Minimum connectivity distance.
        ignore (list or str, optional):
            If not None, includes the names of bath spins which are ignored in the cluster generation.

        strong (bool): Whether to find only completely interconnected clusters (default False).

        nclusters (dict): Dictionary which contain maximum number of clusters of the given size.
            Has the form ``n_clusters = {order: number}``, where ``order`` is the size of the cluster,
            ``number`` is the maximum number of clusters with this size.

            If provided, sorts the clusters by the strength of cluster interaction,
            equal to the lowest pairwise interaction in the cluster. Then the strongest ``number`` of clusters is
            taken.

    Returns:

        dict:
            Dictionary with keys corresponding to size of the cluster,
            and value corresponds to ndarray of shape (matrix, N).
            Here matrix is the number of clusters of given size, N is the size of the cluster.
            Each row contains indexes of the bath spins included in the given cluster.
    """
    graph = make_graph(bath, r_dipole, r_inner=r_inner, ignore=ignore, max_size=5000)
    n_components, labels = connected_components(csgraph=graph, directed=False, return_labels=True)
    if nclusters is None:
        clusters = find_subclusters(order, graph, labels, n_components, strong=strong)
    else:
        clusters = find_valid_subclusters(graph, order, strong=strong, nclusters=nclusters, bath=bath)

    if ignore is not None and order > 0:
        if isinstance(ignore, (str, np.str)):
            clusters[1] = clusters[1][bath[clusters[1]]['N'] != ignore].reshape(-1, 1)
        else:
            for n in ignore:
                clusters[1] = clusters[1][bath[clusters[1]]['N'] != n].reshape(-1, 1)

    return clusters


def make_graph(bath, r_dipole, r_inner=0, ignore=None, max_size=5000):
    """
    Make a connectivity matrix for bath spins.

    Args:
        bath (BathArray): Array of bath spins.
        r_dipole (float): Maximum connectivity distance.
        r_inner (float): Minimum connectivity distance.
        ignore (list or str, optional):
            If not None, includes the names of bath spins which are ignored in the cluster generation.
        max_size (int): Maximum size of the bath before less optimal (but less memory intensive) approach is used.

    Returns:
        crs_matrix: Connectivity matrix.
    """

    if bath.size < max_size:
        dist_matrix = np.linalg.norm(bath['xyz'][:, np.newaxis, :] - bath['xyz'][np.newaxis, :, :], axis=-1)
        atoms_within = np.logical_and(dist_matrix < r_dipole, dist_matrix > r_inner)

    else:
        atoms_within = np.zeros((bath.size, bath.size), dtype=bool)
        for i, a in enumerate(bath):
            dist = np.linalg.norm(bath['xyz'][i:] - a['xyz'], axis=-1)
            atoms_within[i, i:] = (dist < r_dipole) & (dist > r_inner)
    if ignore is not None:
        if isinstance(ignore, (str, np.str)):
            atoms_within = atoms_within & (bath['N'] != ignore)[np.newaxis, :]
        else:
            for n in ignore:
                atoms_within = atoms_within & (bath['N'] != n)[np.newaxis, :]

    if bath.shape[0] == 0:
        print('No spins, no neighbours.')

    # Generate sparse matrix contain connectivity
    graph = csr_matrix(atoms_within, dtype=bool)

    return graph


# Import from connected components from scipy
def connected_components(csgraph, directed=False, connection='weak', return_labels=True):
    """
    Find connected components using ``scipy.sparse.csgraph``.
    See documentation of ``scipy.sparse.csgraph.connected_components``
    """
    return scipy.sparse.csgraph.connected_components(csgraph, directed=directed, connection=connection,
                                                     return_labels=return_labels)


def find_subclusters(maximum_order, graph, labels, n_components, strong=False):
    """
    Find subclusters from connectivity matrix.

    Args:
        maximum_order (int):
            Maximum size of the clusters to find.
        graph (csr_matrix): Connectivity matrix.
        labels (ndarray with shape (n,)): Array of labels of the connected components.
        n_components (int): The number of connected components n.
        strong (bool): Whether to find only completely interconnected clusters (default False).

    Returns:
        dict:
            Dictionary with keys corresponding to size of the cluster,
            and value corresponds to ndarray of shape (matrix, N).
            Here matrix is the number of clusters of given size, N is the size of the cluster.
            Each row contains indexes of the bath spins included in the given cluster.
    """
    # bool 1D array which is true when given element of graph corresponds to
    # cluster component
    clusters = {}
    for k in range(1, maximum_order + 1):
        clusters[k] = []
    # print('Number of disjointed clusters is {}'.format(n_components))
    for component in range(n_components):
        vert_pos = (labels == component)
        vertices = np.nonzero(vert_pos)[0]

        # print('{} cluster contains {} components'.format(component, ncomp))

        # if ncomp <= CCE_order:
        #
        #     clusters[ncomp].append(vertices[np.newaxis, :])
        #
        # else:

        subclusters = {1: vertices[:, np.newaxis]}

        clusters[1].append(vertices[:, np.newaxis])

        if vertices.size >= 2 and maximum_order > 1:
            # Retrieve upper right triangle (remove i,j pairs with i>j),
            # choose only rows corresponding to vertices in the subcluster
            csrmat = scipy.sparse.triu(graph, k=0, format='csr')[vertices]
            # Change to coordinate format of matrix
            coomat = csrmat.tocoo()
            # rows, col give row and colum indexes, which correspond to
            # edges of the graph. as we already slised out the rows,
            # to obtain correct row indexes we need to use vertices array
            row_ind, col_ind = vertices[coomat.row], coomat.col

            bonds = np.column_stack([row_ind, col_ind])
            subclusters[2] = bonds
            clusters[2].append(bonds)

            # Check if [1,2] row in a matrix(Nx2):  any(np.equal(a, [1, 2]).all(1))

            for order in range(3, maximum_order + 1):

                # General way to compute clusters for any order >= 3
                # but for simplicity consider CCE4

                # List of cluster of size 4
                ltriplets = []

                # For ith triplet direct i+1:N pairs, if one of them contains
                # one and only one element of jth pair, they form a cluster of 4
                # There is no need to direct the last one, as it would be included
                # into quartet already if it were to be a part of one
                for i in range(subclusters[order - 1].shape[0] - 1):

                    # The triplet under study
                    test = subclusters[order - 1][i]

                    # For cluster i,j,k (i>j>k, as all indexes are stored in increasing order)
                    # consider only bonds l, n with l >= i, n >= j without loss of generality
                    testbonds = bonds[np.all(bonds >= test[:2], axis=1)]

                    # cond is an bool 2D array of shape (testbonds.shape[0], test.size)
                    # i.e. number of rows corresponds to number of testbonds,
                    # lenght of the row is equal to the length of the test cluster (3 in case CCE4)
                    # cond[i,j] is True if bond[i] contains element of test[j], otherwise False

                    # To construct this array the following procedure is applied:
                    # Reshape testbonds from (n, 2) to (n, 2, 1)
                    # when asked to do logical operation == testbonds is broadcasted to shape (n, 2, order - 1)
                    # In the case of CCE4 (n, 2, 3). Resulting 3D bool array has True entry i,j,k
                    # If j element of testbonds[i] is equal to k element of test
                    # Applying logical operation any along 2nd axis (axis=1, any element of the bond i)
                    # we obtain resulting array cond

                    cond = np.any(testbonds.reshape(testbonds.shape + (1,)) == test, axis=1)
                    # Check which of testbonds form a cluster with the triplet i,j,k
                    # rows is 1D bool array, rows[i] is True if bond[i] contains exactly 1 element of
                    # test triplet
                    rows = np.equal(np.count_nonzero(cond, axis=1), 1)
                    # Prepare 2D array with nrows = number of rows with nonzero entry,
                    # ncols = length of test cluster (for CCE4 is 3)
                    tiled_test = np.tile(test, (np.count_nonzero(rows), 1))

                    if tiled_test.shape[-1] > 2:
                        # From test indexes for each row[i] of nonzero rows choose those indexes, which are not
                        # present in the bond[i],given by reverse cond array
                        flatten = tiled_test[~cond[rows]]
                        # Obtaining correct indexes from tiled test gives flattened array
                        # which should be reshaped nack into (nrows, order - bond).
                        # For CCE4 we need to add 2 indexes
                        # to bond to create a quartet, therefore appendix should have shape (nrows, 2)
                        appendix = flatten.reshape(flatten.size // (order - 2), order - 2)
                    else:
                        # For CCE3 it's easier to do in this way
                        # (probably, idk, I really just don't want to break it)
                        appendix = tiled_test[~cond[rows]][:, np.newaxis]

                    triplets = np.concatenate((testbonds[rows], appendix), axis=1)

                    # If strong keyword was used, the program will find only the completely interconnected clusters
                    # For CCE4 this means that from the given triplet i,j,k to form an interconnected array
                    # i,j,k,l, vertex l should have edges il, jl, kl. Therefore the quartet will appear 3 times
                    # in the array triplets. we choose unique quartets, and from them choose only quartets that
                    # appeared 3 times.
                    if strong and triplets.any():
                        unique, counts = np.unique(np.sort(triplets, axis=1), axis=0, return_counts=True)
                        triplets = unique[counts == order - 1]

                        if triplets.any():
                            ltriplets.append(triplets)
                            # print(triplets)

                    else:
                        ltriplets.append(triplets)

                # Transform list of numpy arrays into numpy array
                try:
                    ltriplets = np.concatenate(ltriplets, axis=0)
                    ltriplets = np.unique(np.sort(ltriplets, axis=1), axis=0)
                except ValueError:
                    break

                subclusters[order] = ltriplets

                clusters[order].append(subclusters[order])

    for o in range(1, maximum_order + 1):
        if clusters[o]:
            # print(clusters[o])
            clusters[o] = np.concatenate(clusters[o], axis=0)
        else:
            print('Set of clusters of order {} is empty!'.format(o))
            clusters.pop(o)

    return clusters


def combine_clusters(cs1, cs2):
    """
    Combine two dictionaries with clusters.

    Args:
        cs1 (dict): First cluster dictionary with keys corresponding to size of the cluster,
            and value corresponds to ndarray of shape (matrix, N).

        cs2 (dict): Second cluster dictionary with the same structure.

    Returns:
        dict: Combined dictionary with unique clusters from both dictionaries.
    """
    keys_1 = list(cs1.keys())
    keys_2 = list(cs2.keys())
    keys = {*keys_1, *keys_2}
    cs_combined = {}
    for k in keys:
        if k in keys_1 and k in keys_2:
            indexes = np.concatenate((cs1[k], cs2[k]))
            cs_combined[k] = np.unique(np.sort(indexes, axis=1), axis=0)
        elif k in keys_1:
            cs_combined[k] = cs1[k]
        elif k in keys_2:
            cs_combined[k] = cs2[k]
    return cs_combined


def expand_clusters(sc):
    """
    Expand dict so each new cluster will include all possible additions of one more bath spin. This increases
    maximum size of the cluster by one.

    Args:
        sc (dict): Initial clusters dictionary.

    Returns:
        dict: Dictionary with expanded clusters.
    """

    indexes = np.arange(sc[1].size, dtype=np.int32)
    comb = np.array([*combinations(indexes, 2)], dtype=np.int32)

    newsc = {}
    newsc[1] = indexes[:, np.newaxis]
    newsc[2] = comb

    for o in sorted(sc)[1:]:
        lexpanded = []

        for test in sc[o]:
            cond = np.any(comb.reshape(comb.shape + (1,)) == test, axis=1)
            rows = np.equal(np.count_nonzero(cond, axis=1), 1)

            tiled_test = np.tile(test, (np.count_nonzero(rows), 1))

            flatten = tiled_test[~cond[rows]]
            appendix = flatten.reshape(-1, o - 1)

            triplets = np.concatenate((comb[rows], appendix), axis=1)
            lexpanded.append(triplets)

        lexpanded = np.concatenate(lexpanded, axis=0)
        lexpanded = np.unique(np.sort(lexpanded, axis=1), axis=0)
        newsc[o + 1] = lexpanded

    return newsc


def find_valid_subclusters(graph, maximum_order, nclusters=None, bath=None, strong=False, ):
    """
    Find subclusters from connectivity matrix.

    Args:
        maximum_order (int):
            Maximum size of the clusters to find.
        graph (csr_matrix): Connectivity matrix.
        nclusters (dict): Dictionary which contain maximum number of clusters of the given size.
        bath (BathArray): Array of bath spins.
        strong (bool): Whether to find only completely interconnected clusters (default False).

    Returns:
        dict:
            Dictionary with keys corresponding to size of the cluster,
            and value corresponds to ndarray of shape (matrix, N).
            Here matrix is the number of clusters of given size, N is the size of the cluster.
            Each row contains indexes of the bath spins included in the given cluster.
    """

    clusters = {1: np.arange(graph.shape[0])[:, np.newaxis]}

    if nclusters is not None and isinstance(nclusters, int):
        nclusters = {k: nclusters for k in range(1, maximum_order + 1)}

    if maximum_order > 1:
        strength = {}
        # Retrieve upper right triangle (remove i,j pairs with i>j),
        csrmat = scipy.sparse.triu(graph, k=0, format='csr')
        # Change to coordinate format of matrix
        coomat = csrmat.tocoo()
        row_ind, col_ind = coomat.row, coomat.col

        bonds = np.column_stack([row_ind, col_ind])

        if nclusters is not None:
            r = bath[col_ind].dist(bath[row_ind])
            # cos_theta = (bath[col_ind].z - bath[row_ind].z) / r
            # strength[2] = np.abs(bath[col_ind].gyro * bath[row_ind].gyro / r ** 3)  # * (1 - 2 * cos_theta ** 2))
            gyros_1 = bath[col_ind].gyro
            gyros_2 = bath[row_ind].gyro
            if len(gyros_1.shape) > 1:
                gyros_1 = np.abs(gyros_1.reshape(gyros_1.shape[0], -1)).max(axis=1)
            if len(gyros_2.shape) > 1:
                gyros_2 = np.abs(gyros_2.reshape(gyros_2.shape[0], -1)).max(axis=1)

            strength[2] = 1 / np.abs(gyros_1 * gyros_2 / r ** 3)  # * (1 - 2 * cos_theta ** 2))
            ordered = strength[2].argsort()  # smallest strength - largest coupling # [::-1] if strength = np.abs
            bonds = bonds[ordered]
            strength[2] = strength[2][ordered]

            if 2 in nclusters:
                bonds = bonds[:nclusters[2]]
                strength[2] = strength[2][:nclusters[2]]

        clusters[2] = bonds

        del coomat, row_ind, col_ind

        for order in range(3, maximum_order + 1):

            ltriplets = []
            # list of triplet strength (used only when n_clusters is not None)
            ltstr = []

            for i in range(clusters[order - 1].shape[0] - 1):

                # The triplet under study
                test = clusters[order - 1][i]
                tripletstrength = None

                # For cluster i,j,k (i>j>k, as all indexes are stored in increasing order)
                # consider only bonds l, n with l >= i, n >= j without loss of generality
                choosebonds = np.all(bonds >= test[:2], axis=1)
                testbonds = bonds[choosebonds]

                cond = np.any(testbonds.reshape(testbonds.shape + (1,)) == test, axis=1)
                # Check which of testbonds form a cluster with the triplet i,j,k
                # rows is 1D bool array, rows[i] is True if bond[i] contains exactly 1 element of
                # test triplet
                rows = np.equal(np.count_nonzero(cond, axis=1), 1)
                # Prepare 2D array with nrows = number of rows with nonzero entry,
                # ncols = length of test cluster (for CCE4 is 3)
                tiled_test = np.tile(test, (np.count_nonzero(rows), 1))

                if tiled_test.shape[-1] > 2:
                    flatten = tiled_test[~cond[rows]]
                    appendix = flatten.reshape(flatten.size // (order - 2), order - 2)
                else:
                    appendix = tiled_test[~cond[rows]][:, np.newaxis]

                triplets = np.concatenate((testbonds[rows], appendix), axis=1)

                if nclusters is not None:
                    teststrength = strength[order - 1][i]
                    tripletstrength = strength[2][choosebonds][rows]
                    # tripletstrength[tripletstrength > teststrength] = teststrength
                    tripletstrength += teststrength

                if strong and triplets.any():
                    unique, index, counts = np.unique(np.sort(triplets, axis=1), axis=0, return_index=True,
                                                      return_counts=True)

                    triplets = unique[counts == order - 1]

                    if triplets.any():
                        ltriplets.append(triplets)

                        if nclusters is not None:
                            tripletstrength = tripletstrength[index[counts == order - 1]]
                            ltstr.append(tripletstrength)

                else:
                    ltriplets.append(triplets)
                    if nclusters is not None:
                        ltstr.append(tripletstrength)

            # Transform list of numpy arrays into numpy array

            try:

                ltriplets = np.concatenate(ltriplets, axis=0)

                # First order by lowest strength, so from two identical triplets
                # one with lower strength will be first in np.unique call

                if nclusters is not None:
                    ltstr = np.concatenate(ltstr)
                    ordered_by_lowest_strength = ltstr.argsort()  # [::-1]  # reverse indexes b/c strength is 1/strength
                    # ordered_by_lowest_strength = ltstr.argsort()
                    ltriplets = ltriplets[ordered_by_lowest_strength]
                    ltstr = ltstr[ordered_by_lowest_strength]

                ltriplets, indexes = np.unique(np.sort(ltriplets, axis=1), axis=0, return_index=True)

                if nclusters is not None:
                    ltstr = ltstr[indexes]
                    ordered_by_strength = ltstr.argsort()  # 0 strength - all bonds are strongly coupled
                    # ordered_by_strength = ltstr.argsort()[::-1]

                    ltriplets = ltriplets[ordered_by_strength]
                    ltstr = ltstr[ordered_by_strength]

                    if order in nclusters:
                        ltstr = ltstr[:nclusters[order]]
                        ltriplets = ltriplets[:nclusters[order]]

            except ValueError:
                print('Set of clusters of order {} is empty!'.format(order))
                break

            clusters[order] = ltriplets
            if nclusters is not None:
                strength[order] = ltstr

    return clusters



from distutils.core import setup

import setuptools

setup(
    name='pycce',
    version='1.0.1',
    url='',
    license='',
    author='Nikita Onizhuk',
    author_email='onizhuk@uchicago.edu',
    description='A package to compute spin dynamics using CCE method',
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy', 'scipy', 'ase', 'pandas'
    ],
    include_package_data=True,
    package_data={'': ['bath/*.txt']},
)



import operator

import numpy as np
import pycce.center
from pycce import dimensions_spinvectors
from pycce.h import total_hamiltonian, conditional_hyperfine, \
    Hamiltonian, bath_hamiltonian
from pycce.run.base import RunObject, simple_propagator, generate_initial_state

_rows = None
_cols = None
_newcols = None


# def _gen_indexes(size):
#     r = np.arange(size)
#     global _rows
#     global _cols
#     global _newcols
#
#     _rows = np.concatenate([r[:i] for i in range(r.size, 0, -1)])
#     _cols = np.concatenate([r[i:] for i in range(r.size, )])
#     _newcols = np.concatenate([np.ones(i, dtype=int) * (r.size - i) for i in range(r.size, 0, -1)])
#
#     if equispaced:
#         if _rows is None:
#             _gen_indexes(corr.size)
#
#         density_matrix = np.zeros((corr.size, corr.size), dtype=np.complex128)
#         density_matrix[_rows, _newcols] = np.triu(corr[np.newaxis, :] - corr[:, np.newaxis])[_rows, _cols]
#         top = density_matrix.sum(axis=0)
#         bottom = np.count_nonzero(density_matrix, axis=0)
#         bottom[bottom == 0] = 1
#         corr = top / bottom

def correlation_it_j0(operator_i, operator_j, dm0_expanded, U):
    """
    Function to compute correlation function of the operator i at time t and operator j at time 0

    Args:
        operator_i (ndarray with shape (n, n)):
            Matrix representation of operator i.
        operator_j (ndarray with shape (n, n)):
            Matrix representation of operator j.
        dm0_expanded (ndarray with shape (n, n)):
            Initial density matrix of the cluster.
        U (ndarray with shape (t, n, n)):
            Time evolution propagator, evaluated over t time points.

    Returns:
        ndarray with shape (t,): Autocorrelation of the z-noise at each time point.

    """

    operator_i_t = np.matmul(U.conj().transpose(0, 2, 1), np.matmul(operator_i, U))
    # operator_j_t = np.matmul(np.transpose(U.conj(), axes=(0, 2, 1)), np.matmul(operator_j, U))
    it_j0 = (np.matmul(operator_i_t, operator_j) + np.matmul(operator_j, operator_i_t)) / 2
    matmul = np.matmul(it_j0, dm0_expanded)
    corr = matmul.trace(axis1=1, axis2=2, dtype=np.complex128)

    return corr


def compute_correlations(nspin, dm0_expanded, U, central_spin=None):
    """
    Function to compute correlations for the given cluster, given time propagator U.

    Args:
        nspin (BathArray):
            BathArray of the given cluster of bath spins.
        dm0_expanded (ndarray with shape (n, n)):
            Initial density matrix of the cluster.
        U (ndarray with shape (t, n, n)):
            Time evolution propagator, evaluated over t time points.
        central_spin (CenterArray): Array of central spins.

    Returns:
        ndarray with shape (t,):
            correlation of the Overhauser field, induced by the given cluster at each time point.

    """
    a_is = np.zeros((3, *U.shape[1:]), dtype=np.complex128)

    dimensions, vectors = dimensions_spinvectors(nspin, central_spin=central_spin)
    if central_spin is not None and central_spin.size > 1:
        raise ValueError('Correlation calculations are supported only for single central spin')

    if len(nspin.A.shape) == len(nspin.shape) + 2:
        for j, n in enumerate(nspin):
            ivec = vectors[j]
            hyperfine_tensor = n['A']
            aivec = np.array([hyperfine_tensor[0, 0] * ivec[0],
                              hyperfine_tensor[1, 1] * ivec[1],
                              hyperfine_tensor[2, 2] * ivec[2]])
            # aivec = np.einsum('ij,jkl->ikl', hyperfine_tensor, ivec)
            # Still don't understand why this doesn't work
            a_is += aivec

        # AI_x = correlation_it_j0(AIs[0], AIs[0], dm0_expanded, U)
        # AI_y = correlation_it_j0(AIs[1], AIs[1], dm0_expanded, U)
        AI_z = correlation_it_j0(a_is[2], a_is[2], dm0_expanded, U)

    else:

        AI_z = []

        for i in range(nspin.A.shape[len(nspin.shape)]):
            for j, n in enumerate(nspin):
                ivec = vectors[j]
                hyperfine_tensor = n['A'][i]
                aivec = np.array([hyperfine_tensor[0, 0] * ivec[0],
                                  hyperfine_tensor[1, 1] * ivec[1],
                                  hyperfine_tensor[2, 2] * ivec[2]])
                a_is += aivec

            AI_z.append(correlation_it_j0(a_is[2], a_is[2], dm0_expanded, U))

    return AI_z  # np.array([AI_x, AI_y, AI_z])


class gCCENoise(RunObject):
    """
    Class for running generalized CCE simulations of the noise autocorrelation function.

    .. note::

        Subclass of the ``RunObject`` abstract class.

    Args:
        *args: Positional arguments of the ``RunObject``.
        **kwargs: Keyword arguments of the ``RunObject``.

    """
    result_operator = operator.iadd
    """Overridden operator which will combine the result of expansion: ``operator.iadd``."""
    contribution_operator = operator.imul
    """Overridden operator which will combine multiple contributions of the same cluster 
    in the optimized approach: ``operator.imul``."""

    removal_operator = operator.isub
    """Overridden operator which remove subcluster contribution
    from the given cluster contribution: ``operator.isub``."""

    addition_operator = np.sum
    """Overridden group operation which will combine contributions from the different clusters into
    one contribution in the direct approach: ``numpy.sum``."""

    def __init__(self, *args, **kwargs):
        self.dm0 = None
        super().__init__(*args, **kwargs)

    def preprocess(self):
        super().preprocess()

        self.dm0 = np.tensordot(self.center.state, self.center.state, axes=0)

    def postprocess(self):
        pass

    def generate_hamiltonian(self):
        """
        Using the attributes of the ``self`` object,
        compute the cluster hamiltonian including the central spin.

        Returns:
            Hamiltonian: Cluster hamiltonian.

        """
        ham = total_hamiltonian(self.cluster, self.center, self.magnetic_field)
        # ham.data += zero_order_addition(ham.vectors, self.cluster, self.others, self.others.proj)

        return ham

    def compute_result(self):
        """
        Using the attributes of the ``self`` object,
        compute autocorrelation function of the noise from bath spins in the given cluster.

        Returns:

            ndarray: Computed autocorrelation function.
        """
        time_propagator = simple_propagator(self.timespace, self.hamiltonian)

        dmtotal0 = generate_initial_state(self.base_hamiltonian.dimensions, central_state=self.dm0, states=self.states)

        return compute_correlations(self.cluster, dmtotal0, time_propagator, central_spin=self.center)


class CCENoise(RunObject):
    """
    Class for running conventional CCE simulations of the noise autocorrelation function.

    .. note::

        Subclass of the ``RunObject`` abstract class.


    .. warning::

        In general, for calculations of the autocorrelation function, better results are achieved with
        generalized CCE, which accounts for the evolution of the entangled state of the central spin.

        Second order couplings between nuclear spins are not implemented.

    Args:
        *args: Positional arguments of the ``RunObject``.
        **kwargs: Keyword arguments of the ``RunObject``.

    """
    result_operator = operator.iadd
    """Overridden operator which will combine the result of expansion: ``operator.iadd``."""
    contribution_operator = operator.imul
    """Overridden operator which will combine multiple contributions of the same cluster 
    in the optimized approach: ``operator.imul``."""

    removal_operator = operator.isub
    """Overridden operator which remove subcluster contribution
    from the given cluster contribution: ``operator.isub``."""

    addition_operator = np.sum
    """Overridden group operation which will combine contributions from the different clusters into
    one contribution in the direct approach: ``numpy.sum``."""

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.projections = None

    def preprocess(self):
        super().preprocess()
        pycce.center.generate_projections()
        if self.center.size > 1:
            raise ValueError('Correlation calculations are supported only for single central spin')

        self.projections = (np.abs(self.center.projections_alpha) + np.abs(self.center.projections_beta)) / 2
        # self.projections = generate_projections(self.state)

    def postprocess(self):
        global _cols
        global _rows
        global _newcols

        _cols = None
        _rows = None
        _newcols = None

    def generate_hamiltonian(self):
        """
        Using the attributes of the ``self`` object,
        compute the projected cluster hamiltonian, averaged for two qubit states.

        Returns:
            Hamiltonian: Cluster hamiltonian.

        """
        totalh = bath_hamiltonian(self.cluster, self.magnetic_field)

        for ivec, n in zip(totalh.vectors, self.cluster):
            hsingle = 0
            for i, proj in enumerate(self.projections):
                if self.center.size > 1:
                    hf = n['A'][i]
                else:
                    hf = n['A']

                hsingle += conditional_hyperfine(hf, ivec, proj)

            totalh.data += hsingle

        # totalh.data += zero_order_addition(totalh.vectors, self.cluster, self.others, self.others.proj)

        return totalh
        # ham = total_hamiltonian(self.cluster, self.magnetic_field, self.zfs, others=self.others,
        #                         other_states=self.other_states, central_gyro=self.gyro, central_spin=self.spin)
        # return ham

    def compute_result(self):
        """
        Using the attributes of the ``self`` object,
        compute autocorrelation function of the noise from bath spins in the given cluster.

        Returns:

            ndarray: Computed autocorrelation function.
        """
        dm0_expanded = generate_initial_state(self.base_hamiltonian.dimensions, states=self.states, central_state=None)

        time_propagator = simple_propagator(self.timespace, self.hamiltonian)

        return compute_correlations(self.cluster, dm0_expanded, time_propagator)



import warnings

import numpy as np
from numba import jit

from .array import BathArray, check_gyro
from ..constants import BOHR_TO_ANGSTROM, HBAR_MU0_O4PI, ELECTRON_GYRO, PI2

# Copied from ASE
chemical_symbols = [
    # 0
    'X',
    # 1
    'H', 'He',
    # 2
    'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne',
    # 3
    'Na', 'Mg', 'Al', 'Si', 'P', 'spin_matrix', 'Cl', 'Ar',
    # 4
    'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn',
    'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr',
    # 5
    'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd',
    'In', 'Sn', 'Sb', 'Te', 'I', 'Xe',
    # 6
    'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy',
    'Ho', 'Er', 'Tm', 'Yb', 'Lu',
    'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi',
    'Po', 'At', 'Rn',
    # 7
    'Fr', 'Ra', 'Ac', 'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk',
    'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr',
    'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg', 'Cn', 'Nh', 'Fl', 'Mc',
    'Lv', 'Ts', 'Og']


class Cube:
    """
    Class to process the .cube datafiles with spin polarization.

    Args:
        filename (str): Name of the .cube file.

    Attributes:
        comments (str): First two lines of the .cube file.
        origin (ndarray with shape (3,)): Coordinates of the origin in angstrom.
        voxel (ndarray with shape (3,3)): Parameters of the voxel - unit of the 3D grid in angstrom.
        size (ndarray with shape (3,)): Size of the cube.
        atoms (BathArray with shape (n)): Array of atoms in the cube.
        data (ndarray with shape (size[0], size[1], size[2]): Data stored in cube.
        grid (ndarray with shape (size[0], size[1], size[2], 3): Coordinates of the points at which data is computed.
        integral (float): Data integrated over cube.
        spin (float): integral / 2 - total spin.
    """
    _dt = np.dtype([('N', np.unicode_, 16), ('xyz', np.float64, (3,))])

    def __init__(self, filename):

        with open(filename, "r") as content:
            # first two lines are comments
            self.comments = next(content).strip() + "\n" + next(content).strip()

            # total number of bath | xyz of the cube origin
            tot = next(content).split()
            natoms = int(tot[0])

            self.origin = np.array([float(x) for x in tot[1:]]) * BOHR_TO_ANGSTROM
            self.voxel = np.empty([3, 3], dtype=np.float64)
            self.size = np.empty(3, dtype=np.int32)

            self.atoms = BathArray((natoms,))

            for i in range(3):
                tot = next(content).split()
                self.size[i] = int(tot[0])

                if self.size[i] < 0:
                    self.voxel[i] = [float(x) for x in tot[1:]]

                else:
                    self.voxel[i] = [float(x) * BOHR_TO_ANGSTROM for x in tot[1:]]
            with warnings.catch_warnings(record=True):
                for j in range(natoms):
                    tot = next(content).split()
                    self.atoms[j]['N'] = chemical_symbols[int(tot[0])]
                    self.atoms[j]['xyz'] = [float(x) for x in tot[2:]]

            if self.size[0] > 0:
                self.atoms['xyz'] *= BOHR_TO_ANGSTROM

            data = [float(x) for line in content for x in line.split()]

        if self.size[0] < 0:
            self.data = np.array(data).reshape(np.abs(self.size))
        else:
            self.data = np.array(data).reshape(np.abs(self.size)) / (BOHR_TO_ANGSTROM ** 3)

        # detect if diagonal
        # (see https://stackoverflow.com/questions/43884189/check-if-a-large-matrix-is-diagonal-matrix-in-python)
        if np.any(self.voxel.reshape(-1)[:-1].reshape(2, 4)[:, 1:]):
            warnings.warn('Voxel might be non-orthogonal. Correctness of the results is not tested')

        na = np.newaxis
        a = np.arange(self.size[0])[:, na] * self.voxel[0][na, :]
        b = np.arange(self.size[1])[:, na] * self.voxel[1][na, :]
        c = np.arange(self.size[2])[:, na] * self.voxel[2][na, :]
        mesh = a[:, na, na, :] + b[na, :, na, :] + c[na, na, :, :]

        self.grid = mesh + self.origin
        self.integral = np.trapz(np.trapz(np.trapz(self.data))) * np.linalg.det(self.voxel)
        self.spin = round(self.integral) * 0.5

    def transform(self, rotmatrix=None, shift=None):
        r"""
        Changes coordinates of the grid. DOES NOT ASSUME PERIODICITY.

        Args:
            rotmatrix (ndarray with shape (3, 3)): Rotation matrix `R`:

                .. math::

                    R =  &[n_1^{(1)} n_1^{(2)} n_1^{(3)}]\\
                         &[n_2^{(1)} n_2^{(2)} n_2^{(3)}]\\
                         &[n_3^{(1)} n_3^{(2)} n_3^{(3)}]

                where :math:`n_i^{(j)}` corresponds to the coefficient of initial basis vector :math:`i`
                for :math:`j` new basis vector:

                .. math::

                    e'_j = n_1^{(j)} \vec{e}_1 + n_2^{(j)} \vec{e}_2 + n_3^{(j)} \vec{e}_3

                In other words, columns of `R` are coordinates of the new
                basis in the old basis.

                Given vector in initial basis v = [v1, v2, v3],
                vector in new basis is given as v' = R.T @ v.

            shift (ndarray with shape (3,)): Shift in the origin of coordinates (in the old basis).

        """

        if shift is not None:
            shift = np.asarray(shift)
            self.grid = self.grid + shift

            self.atoms.xyz += shift
            self.origin += shift

        if rotmatrix is not None:
            assert (np.isclose(np.linalg.det(rotmatrix), 1.)), 'Determinant of R is not equal to 1'
            self.grid = np.einsum('ij,abcj->abci', rotmatrix.T, self.grid)
            self.origin = rotmatrix.T @ self.origin
            self.atoms.transform(rotation_matrix=rotmatrix)
        return

    def integrate(self, position, gyro_n, gyro_e=ELECTRON_GYRO, spin=None, parallel=False, root=0):
        """
        Integrate over polarization data, stored in Cube object,
        to obtain hyperfine dipolar-dipolar tensor.

        Args:
            position (ndarray with shape (3,) or (n, 3)): Position of the bath spin at which to compute
                hyperfine tensor or array of positions.
            gyro_n (float or ndarray with shape (n,) ): Gyromagnetic ratio of the bath spin or array of the ratios.
            gyro_e (float): Gyromagnetic ratio of central spin.
            spin (float): Total spin of the central spin. If not given, taken from the integral of the polarization.

        Returns:
            ndarray with shape (3, 3) or (n, 3, 3): Hyperfine tensor or array of hyperfine tensors.
        """
        if parallel:
            try:
                import mpi4py
            except ImportError:
                warnings.warn('Could not find mpi4py. Using the serial implementation instead')
                parallel = False

        if parallel:
            comm = mpi4py.MPI.COMM_WORLD

            size = comm.Get_size()
            rank = comm.Get_rank()

        else:
            rank = root

        if spin is None:
            spin = self.spin

        if np.around(spin * 2) != np.around(self.integral):
            warnings.warn(f'provided spin: {spin:.2f} is not equal to one from spin density: {self.integral / 2:.2f}')

        if rank == root:
            gyro_e, _ = check_gyro(gyro_e)
            gyro_n, _ = check_gyro(gyro_n)

            gyro_e = np.asarray(gyro_e)
            gyro_n = np.asarray(gyro_n)
            position = np.asarray(position)

        if parallel:
            gyro_e = comm.bcast(gyro_e, root)
            gyro_n = comm.bcast(gyro_n, root)
            position = comm.bcast(position, root)

        if len(position.shape) <= 1:
            return _cube_integrate(self.data, self.grid, self.voxel, spin,
                                   position, gyro_n, gyro_e)

        npos = position.shape[0]
        nshape = (npos, *gyro_n.shape[1:])
        gyro_n = np.array(np.broadcast_to(gyro_n, nshape))
        if gyro_n.ndim < 3:
            gyro_n = gyro_n[:, np.newaxis, ...]
        if parallel:
            start, end = _distribute(rank, size, npos)
            position = position[start:end]
            gyro_n = gyro_n[start:end]

        if position.size:
            hyperfine = _cube_integrate_array(self.data, self.grid, self.voxel, spin, position,
                                              gyro_n, gyro_e)
        else:
            hyperfine = np.zeros((0, 3, 3), dtype=np.float64)

        if parallel:

            total_hyperfine = np.zeros((npos * 3 * 3), dtype=np.float64)

            start_ends = tuple(_distribute(i, size, npos) for i in range(size))

            sendcounts = tuple((end - start) * 3 * 3 for (start, end) in start_ends)
            displacements = tuple(start * 3 * 3 for (start, end) in start_ends)

            comm.Allgatherv([hyperfine.flatten(), mpi4py.MPI.DOUBLE],
                            [total_hyperfine, sendcounts, displacements, mpi4py.MPI.DOUBLE])

            total_hyperfine = total_hyperfine.reshape(npos, 3, 3)
            comm.Barrier()

        else:
            total_hyperfine = hyperfine

        return total_hyperfine

        # d['A'] = -(3 * np.outer(pos, pos) - identity * r ** 2) / (r ** 5) * pre


def _distribute(rank, size, number):
    remainder = number % size
    add = int(rank < remainder)
    each = number // size
    start = rank * each + rank if rank < remainder else rank * each + remainder
    end = start + each + add

    return start, end


@jit(nopython=True)
def _cube_integrate(data, grid, voxel, spin, position, gyro_n, gyro_e=ELECTRON_GYRO):
    """
    Integrate cube for one position.

    Args:
        data (ndarray): 3D data.
        grid (ndarray): Grid over which to integrate.
        voxel (ndarray): Parameters of the voxel.
        spin (float): Total central spin.
        position (ndarray with shape (3,)): Position of the bath spin at which to compute
            hyperfine tensor
        gyro_n (ndarray): Gyromagnetic ratio of the bath spin.
        gyro_e (ndarray): Gyromagnetic ratio of central spin.

    Returns:
        ndarray with shape (3, 3): Hyperfine tensor.
    """
    pos = grid - position

    dist = np.sqrt(np.sum(pos ** 2, axis=-1))

    hyperfine = np.zeros((3, 3), dtype=np.float64)
    for i in range(3):
        for j in range(3):

            if i == j:
                integrand = - data * (3 * pos[:, :, :, i] * pos[:, :, :, j] - dist ** 2) / dist ** 5
            else:
                integrand = - data * (3 * pos[:, :, :, i] * pos[:, :, :, j]) / dist ** 5

            hyperfine[i, j] = np.trapz(np.trapz(np.trapz(integrand))) * HBAR_MU0_O4PI / (
                    2 * spin * PI2) * np.linalg.det(voxel)

    if gyro_e.ndim < 2:
        hyperfine = hyperfine * gyro_e
    else:
        hyperfine = gyro_e @ hyperfine

    if gyro_n.ndim < 2:
        hyperfine = hyperfine * gyro_n
    else:
        hyperfine = hyperfine @ gyro_n

    return hyperfine


@jit(nopython=True)
def _cube_integrate_array(data, grid, voxel, spin, coordinates, gyros, gyro_e=ELECTRON_GYRO):
    """
    Integrate cube for array of positions.
    Args:
        data (ndarray): 3D data.
        grid (ndarray): Grid over which to integrate.
        voxel (ndarray): Parameters of the voxel.
        spin (float): Total central spin.
        coordinates (ndarray with shape (n, 3)): Positions of the bath spins at which to compute
                hyperfine tensors.
        gyros (ndarray with shape (n,) ): Array of gyromagnetic ratio of the bath spins.
        gyro_e (ndarray): Gyromagnetic ratio of central spin.

    Returns:
        ndarray with shape (n, 3, 3): Array of hyperfine tensors.
    """
    hyperfines = np.zeros((coordinates.shape[0], 3, 3), dtype=np.float64)

    for i, position in enumerate(coordinates):
        hyperfines[i] = _cube_integrate(data, grid, voxel, spin, position, gyros[i, ...], gyro_e)

    return hyperfines



import functools

import numpy as np
from numpy import ma as ma
from pycce.utilities import gen_state_list
from numba.typed import List


def generate_bath_state(bath, nbstates, seed=None, parallel=False):
    r"""
    Generator of the random *pure* :math:`\hat{I}_z` bath eigenstates.

    Args:
        bath (BathArray): Array of bath spins.
        nbstates (int): Number of random bath states to generate.
        seed (int): Optional. Seed for RNG.
        parallel (bool): True if run in parallel mode. Default False.

    Yields:
        List: list of the pure bath spin state vectors.
    """
    rgen = np.random.default_rng(seed)
    rank = 0
    comm = None

    if parallel:
        try:
            import mpi4py
            comm = mpi4py.MPI.COMM_WORLD
            rank = comm.Get_rank()

        except ImportError:
            print('Parallel failed: mpi4py is not found. Running serial')
            parallel = False

    dimensions = np.empty(bath.shape, dtype=np.int32)

    for _ in range(nbstates):
        bath_state = np.empty(bath.shape, dtype=np.float64)

        if rank == 0:

            for n in np.unique(bath.N):
                s = bath.types[n].s
                snumber = np.int32(round(2 * s + 1))
                mask = bath.N == n
                bath_state[mask] = rgen.integers(snumber, size=np.count_nonzero(mask)) - s

                dimensions[mask] = snumber

        if parallel:
            comm.Bcast(bath_state, root=0)
            comm.Bcast(dimensions, root=0)

        bath_state = gen_state_list(bath_state, dimensions)

        yield bath_state


def monte_carlo_method_decorator(func):
    """
    Decorator to sample over random bath states given function.
    """

    @functools.wraps(func)
    def inner_method(self, *args,
                     **kwargs):
        seed = self.seed
        if self.parallel_states:
            try:
                from mpi4py import MPI
            except ImportError:
                print('Parallel failed: mpi4py is not found. Running serial')
                self.parallel_states = False

        if isinstance(self.masked, bool) and not self.masked:
            self.masked = None

        if self.masked is not None:
            divider = 0
        else:
            root_divider = self.nbstates

        if self.parallel_states:
            comm = MPI.COMM_WORLD

            size = comm.Get_size()
            rank = comm.Get_rank()

            remainder = self.nbstates % size
            add = int(rank < remainder)
            nbstates = self.nbstates // size + add

            if seed:
                seed = seed + rank * 19653252
        else:
            rank = 0
            nbstates = self.nbstates

        total = 0j

        his = self.bath.state.has_state.copy()  # have_initial_state
        for bath_state in generate_bath_state(self.bath[~his], nbstates, seed=seed, parallel=self.parallel):

            self.bath.state[~his] = bath_state
            result = func(self, *args, **kwargs)

            if self.masked is not None:
                if isinstance(self.masked, bool):
                    self.masked = np.abs(result[0] * 1.01)

                proper = np.abs(result) <= self.masked
                divider += proper.astype(np.int32)
                result[~proper] = 0.

            total += result

        self.bath.state[~his] = None

        if self.parallel_states:
            if rank == 0:
                result_shape = total.shape
            else:
                result_shape = None

            result_shape = comm.bcast(result_shape, root=0)
            total = np.asarray(total)
            if not total.shape:
                total = np.zeros(result_shape,

                                 dtype=np.complex128)
            root_result = ma.array(np.zeros(result_shape), dtype=np.complex128)
            comm.Allreduce(total, root_result, MPI.SUM)

            if self.masked is not None:
                if rank == 0:
                    divider_shape = divider.shape
                else:
                    divider_shape = None

                divider_shape = comm.bcast(divider_shape, root=0)

                root_divider = np.zeros(divider_shape, dtype=np.int32)
                comm.Allreduce(divider, root_divider, MPI.SUM)

        else:
            root_result = total
            if self.masked is not None:
                root_divider = divider

        root_result = ma.array(root_result, fill_value=0, dtype=np.complex128)

        if self.masked is not None:
            root_result[root_divider == 0] = ma.masked
            # root_divider = root_divider.reshape(root_result.shape[0], *[1] * len(root_result.shape[1:]))

        root_result /= root_divider

        return root_result

    return inner_method



import re
import sys
import warnings
from collections import defaultdict
from string import digits

import numpy as np
from pycce.utilities import rotmatrix

from .array import BathArray
from .array import common_concentrations

err_range = 0.1
FLOAT_ERROR_RANGE = 1e-10


class BathCell:
    r"""
    Generator of the bath spins positions from the unit cell of the material.

    Args:
        a (float): `a` parameter of the primitive cell.
        b (float): `b` parameter of the primitive cell.
        c (float): `c` parameter of the primitive cell.
        alpha (float): :math:`\alpha` angle of the primitive cell.
        beta (float): :math:`\beta` angle of the primitive cell.
        gamma (float): :math:`\gamma` angle of the primitive cell.
        angle (str): units of the :math:`\alpha`, :math:`\beta`, :math:`\gamma` angles.
            Can be either radians (``'rad'``), or degrees (``'deg'``).
        cell (ndarray with shape (3, 3)): Parameters of the cell.

            ``cell`` is 3x3 matrix with columns of coordinates of crystallographic vectors
            in the cartesian reference frame. See ``cell`` attribute.

            If provided, overrides `a`, `b`, and `c`.

    Attributes:

        cell (ndarray with shape (3, 3)): Parameters of the cell.
            ``cell`` is 3x3 matrix with entries:

            .. math::

                [&[a_x\ b_x\ c_x]\\
                &[a_y\ b_y\ c_y]\\
                &[a_z\ b_z\ c_z]]

            where a, b, c are crystallographic vectors
            and x, y, z are their coordinates in the cartesian reference frame.

        atoms (dict): Dictionary containing coordinates and occupancy of each lattice site::

                {atom_1: [array([x1, y1, z1]), array([x2, y2, z2])],
                 atom_2: [array([x3, y3, z3]), ...]}

        isotopes (dict):
            Dictionary containing spin types and their concentration for each lattice site type::

                {atom_1: {spin_1: concentration, spin_2: concentration},
                 atom_2: {spin_3: concentration ...}}

            where ``atom_i`` are lattice site types, and ``spin_i`` are spin types.
    """

    _conv = {'rad': 1, 'deg': 2 * np.pi / 360}
    _coord_types = ['angstrom', 'cell']

    def __init__(self, a=None, b=None, c=None,
                 alpha=None, beta=None, gamma=None,
                 angle='rad', cell=None):

        if angle not in self._conv:
            raise KeyError('Only angle available are: '
                           ', '.join(self._conv.keys()))

        if alpha is None:
            alpha = np.pi / 2
        else:
            alpha = alpha * self._conv[angle]

        if beta is None:
            beta = np.pi / 2
        else:
            beta = beta * self._conv[angle]

        if gamma is None:
            gamma = np.pi / 2
        else:
            gamma = gamma * self._conv[angle]
        if b is None:
            b = a
        if c is None:
            c = a
        # self.state, self.beta, self.gamma = state, beta, gamma
        # self.a, self.b, self.c = a, b, c
        if cell is not None:
            self.cell = np.asarray(cell)
        else:
            inbr = (1 + 2 * np.cos(alpha) * np.cos(beta) * np.cos(gamma) -
                    np.cos(alpha) ** 2 - np.cos(beta) ** 2 - np.cos(gamma) ** 2)

            volume = a * b * c * inbr ** (1 / 2)

            self.cell = np.zeros((3, 3))
            self.cell[0, 0] = a
            self.cell[0, 1] = b * np.cos(gamma)
            self.cell[1, 1] = b * np.sin(gamma)
            self.cell[0, 2] = c * np.cos(beta)
            self.cell[1, 2] = c * (np.cos(alpha) - np.cos(beta) *
                                   np.cos(gamma)) / np.sin(gamma)
            # cell is 3x3 matrix with entries:
            # cell = [a_x b_x c_x]
            #        [a_y b_y c_y]
            #        [a_z b_z c_z]

            self.cell[2, 2] = volume / a / b / np.sin(gamma)

        if np.linalg.cond(self.cell) < 1 / sys.float_info.epsilon:
            zdr = self.to_cell([0, 0, 1])
            zdr = zdr / np.linalg.norm(zdr)
        else:
            zdr = np.zeros(3)

        self._zdir = zdr

        self.atoms = defaultdict(list)
        self.isotopes = defaultdict(dict)

    @property
    def zdir(self):
        """
        ndarray: z-direction of the reference cartesian coordinate frame in cell coordinates.
        """
        return self._zdir

    @zdir.setter
    def zdir(self, direction):
        self._zdir = np.asarray(direction) / np.linalg.norm(direction)  # Stored in the cell coordinates

        a = np.array([0, 0, 1])  # Initial vector

        ud = self.cell @ self._zdir
        b = ud / np.linalg.norm(ud)  # Final vector

        # Rotational matrix
        # If z direction is opposite
        if abs(a @ b + 1) < FLOAT_ERROR_RANGE:
            rotation = np.array([[0, 1, 0],
                                 [1, 0, 0],
                                 [0, 0, -1]])

        elif abs(a @ b - 1) < FLOAT_ERROR_RANGE:
            rotation = np.eye(3)

        else:
            # a -> b
            rotation = rotmatrix(a, b)

        self.cell = np.linalg.inv(rotation) @ self.cell
        self.cell[np.abs(self.cell) < FLOAT_ERROR_RANGE * np.max(np.abs(self.cell))] = 0

        # R =  [n_1^(1) n_1^(2) n_1^(3)]
        #      [n_2^(1) n_2^(2) n_2^(3)]
        #      [n_3^(1) n_3^(2) n_3^(3)]
        # n_i^(j) corresponds to coeff of initial basis vector i
        # for j new basis vector:

        # e'_j = n_1^(j)*e_1 + n_2^(j)*e_2 + n_3^(j)*e_3
        # in other words, columns of matrix are coordinates of the new
        # basis in the old basis.

    def rotate(self, rotation_matrix):
        """
        Rotate the BathCell using the rotation matrix provided.

        Args:
            rotation_matrix (ndarray with shape (3,)): Rotation matrix R which rotates the old basis of the
                cartesian reference frame to the new basis.
        """

        self.cell = np.linalg.inv(rotation_matrix) @ self.cell
        self.cell[np.abs(self.cell) < FLOAT_ERROR_RANGE * np.max(np.abs(self.cell))] = 0

        zd = self.to_cell([0, 0, 1])
        self._zdir = zd / np.linalg.norm(zd)
        self._zdir[np.abs(self._zdir) < FLOAT_ERROR_RANGE * np.max(np.abs(self._zdir))] = 0

    def set_zdir(self, direction, type='cell'):
        """
        Set z-direction of the cell.

        Args:
            direction (ndarray with shape (3,)): Direction of the z axis.

            type (str): How coordinates in ``direction`` are stored. If ``type="cell"``,
                assumes crystallographic coordinates. If ``type="angstrom"`` assumes that z direction is given
                in the cartresian reference frame.

        """
        if type == 'cell':
            direction = np.asarray(direction)
        elif type == 'angstrom':
            direction = self.to_cell(np.asarray(direction))
        else:
            raise ValueError('Unknown direction type.')
        self.zdir = direction

    def add_atoms(self, *args, type='cell'):
        """
        Add coordinates of the lattice sites to the unit cell.

        Args:
            *args (tuple): List of tuples, each containing the type of atom N (*str*),
                and the xyz coordinates in the format (*float, float, float*): ``(N, [x, y, z])``.

            type (str): Type of coordinates. Can take values of ``['cell', 'angstrom']``.

                If ``type="cell"``, assumes crystallographic coordinates.

                If ``type="angstrom"`` assumes that coordinates are given in the cartresian reference frame.

        Returns:

            dict:
                View of ``cell.atoms`` dictionary, where each key is the type of lattice site, and each value
                is the list of coordinates in crystallographic frame.

        Examples:

            >>> cell = BathCell(10)
            >>> cell.add_atoms(('C', [0, 0, 0]), ('C', [5, 5, 5]), type='angstrom')
            >>> cell.add_atoms(('Si', [0, 0.5, 0.]), type='cell')
            >>> print(cell.atoms)
            {'C': [array([0., 0., 0.]), array([0.5, 0.5, 0.5])], 'Si': [array([0. , 0.5, 0. ])]}

        """

        for tup in args:
            if type == 'cell':
                coord = np.asarray(tup[1])

            elif type == 'angstrom':
                coord = self.to_cell(tup[1])

            else:
                raise ValueError('Unknown coordinates type. Supported:'
                                 '\n'.join(str(x) for x in self._coord_types))

            if tup[0] in self.atoms:
                self.atoms[tup[0]].append(coord)

            else:
                self.atoms[tup[0]] = [coord]

        return self.atoms

    def add_isotopes(self, *args):
        """
        Add spins that can populate each lattice site type.

        Args:

            *args (tuple or list of tuples): Each tuple can have any of the following formats:

                * Name of the lattice site `N` (`str`), name of the spin `X` (`str`),
                  concentration `c` (`float`, in decimal): ``(N, X, c)``.

                * Isotope name `X and concentration `c`: ``(X, c)``.

                  In this case, the name of the isotope is given in the format
                  ``"{}{}".format(digits, atom_name)`` where ``digits`` is any set of digits 0-9,
                  ``atom_name`` is the name of the corresponding lattice site.
                  Convenient when generating nuclear spin bath.

        Returns:
            dict:
                View of ``cell.isotopes`` dictionary which contains information about lattice site types, spin types,
                and their concentrations::

                    {atom_1: {spin_1: concentration, spin_2: concentration},
                     atom_2: {spin_3: concentration ...}}

        Examples:

            >>> cell = BathCell(10)
            >>> cell.add_atoms(('C', [0, 0, 0]), ('C', [5, 5, 5]), type='angstrom')
            >>> cell.add_isotopes(('C', 'X', 0.001), ('13C', 0.0107))
            >>> print(cell.isotopes)
            {'C': {'X': 0.001, '13C': 0.0107}}

        """

        remove_digits = str.maketrans('', '', digits)

        for tup in args:
            try:

                atom_name = tup[0]
                isotope_name = tup[1]
                concentration = tup[2]

                if atom_name in self.isotopes:
                    self.isotopes[atom_name][isotope_name] = concentration

                else:
                    self.isotopes[atom_name] = {isotope_name: concentration}

            except IndexError:

                isotope_name = tup[0]
                concentration = tup[1]
                atom_name = isotope_name.translate(remove_digits)

                if atom_name in self.isotopes:
                    self.isotopes[atom_name][isotope_name] = concentration
                else:
                    self.isotopes[atom_name] = {isotope_name: concentration}

        return self.isotopes

    def gen_supercell(self, size, add=None, remove=None, seed=None, recenter=True):
        """
        Generate supercell populated with spins.

        .. note::

            If ``isotopes`` were not provided, assumes the natural concentration of nuclear spin isotopes for each
            lattice site type. However, if any isotope concentration is provided,
            then uses only user-defined ones.

        Args:
            size (float): Approximate linear size of the supercell. The generated supercell will have
                minimal distance between opposite sides larger than this parameter.

            add (tuple or list of tuples):
                Tuple or list of tuples containing common_isotopes to add as a defect.
                Each tuple contains name of the new isotope
                and its coordinates in the cell basis: ``(isotope_name, x_cell, y_cell, z_cell)``.

            remove (tuple or list of tuples):
                Tuple or list of tuples containing bath to remove in the defect.
                Each tuple contains name of the atom to remove
                and its coordinates in the cell basis: ``(atom_name, x_cell, y_cell, z_cell)``.

            seed (int): Seed for random number generator.

            recenter (bool):
                True if place approximate center of the supercell at (0,0,0). False if start supercell
                at (0, 0, 0). Default True.

        .. note::

            While ``add`` takes the **spin** name as an argument, ``remove`` takes the lattice site name.

        Returns:
            BathArray: Array of the spins in the given supercell.
        """
        if not self.isotopes:
            isotopes = {}

            for a in self.atoms:

                try:
                    isotopes[a] = common_concentrations[a]
                except KeyError:
                    pass

        else:
            isotopes = self.isotopes

        rgen = np.random.default_rng(seed)

        axb = np.cross(self.cell[:, 0], self.cell[:, 1])
        bxc = np.cross(self.cell[:, 1], self.cell[:, 2])
        cxa = np.cross(self.cell[:, 2], self.cell[:, 0])
        # Number of unit cells along a,b,c directions
        anumber = int(size * np.linalg.norm(bxc) / (bxc @ self.cell[:, 0]) + 1)
        bnumber = int(size * np.linalg.norm(cxa) / (cxa @ self.cell[:, 1]) + 1)
        cnumber = int(size * np.linalg.norm(axb) / (axb @ self.cell[:, 2]) + 1)
        # print(anumber, bnumber, cnumber)

        dt = np.dtype([('N', np.unicode_, 16), ('xyz', np.float64, (3,))])
        atoms = []

        for a in isotopes:
            # Number of sites for given type of atom
            nsites = len(self.atoms[a])
            # Cartesian of those positions
            sites_xyz = np.asarray(self.atoms[a]) @ self.cell.T
            # print(sites_xyz)

            maxind = np.array([anumber,
                               bnumber,
                               cnumber,
                               nsites], dtype=np.int32)

            natoms = np.prod(maxind, dtype=np.int32)
            # 0, 1,2, .. natoms
            atom_seedsites = np.arange(natoms, dtype=np.int32)
            mask = np.zeros(natoms, dtype=bool)
            # C: 13C, 14C
            for i in isotopes[a]:
                conc = isotopes[a][i]
                nisotopes = int(round(natoms * conc))
                seedsites = rgen.choice(atom_seedsites[~mask],
                                        nisotopes, replace=False,
                                        shuffle=False)

                mask += np.isin(atom_seedsites, seedsites)
                bcn = bnumber * cnumber * nsites
                cn = cnumber * nsites

                if recenter:
                    aindexes = seedsites // bcn - (anumber - 1) // 2  # recenter at 0
                    bindexes = (seedsites % bcn) // cn - (bnumber - 1) // 2
                    cindexes = ((seedsites % bcn) % cn) // nsites - (cnumber - 1) // 2
                else:
                    aindexes = seedsites // bcn
                    bindexes = (seedsites % bcn) // cn
                    cindexes = ((seedsites % bcn) % cn) // nsites

                # indexes of the sites
                nindexes = ((seedsites % bcn) % cn) % nsites

                indexes = np.column_stack((aindexes,
                                           bindexes,
                                           cindexes))

                uc_positions = np.einsum('jk,ik->ij', self.cell, indexes)

                subatoms = np.zeros(indexes.shape[0], dtype=dt)
                subatoms['N'] = i
                subatoms['xyz'] = uc_positions + sites_xyz[nindexes]

                atoms.append(subatoms)

        atoms = np.concatenate(atoms)
        # bath = bath[np.linalg.norm(bath['xyz'], axis=1) <= size]

        defective_atoms = defect(self.cell, atoms, add=add, remove=remove)
        bath = BathArray(array=defective_atoms)
        return bath

    def to_cartesian(self, coord):
        """
        Transform coordinates from crystallographic basis to the cartesian reference frame.

        Args:
            coord (ndarray with shape (3,) or (n, 3)): Coordinates in crystallographic basis or array of coordinates.

        Returns:
            ndarray with shape (3,) or (n, 3): Cartesian coordinates in angstrom.
        """

        coord = np.asarray(coord)
        if len(coord.shape) == 1:
            return self.cell @ coord

        elif len(coord.shape) == 2:
            return np.einsum('jk,ik->ij', self.cell, coord)
        else:
            raise ValueError('Improper coordinates format')

    def to_cell(self, coord):
        """
        Transform coordinates from the cartesian coordinates of the reference frame to the cell coordinates.

        Args:
            coord (ndarray with shape (3,) or (n, 3)): Cartesian coordinates in angstrom or array of coordinates.

        Returns:
            ndarray with shape (3,) or (n, 3): Coordinates in the cell basis.

        """
        coord = np.asarray(coord)

        if len(coord.shape) == 1:
            return np.linalg.inv(self.cell) @ coord

        elif len(coord.shape) == 2:
            return np.einsum('jk,ik->ij', np.linalg.inv(self.cell), coord)
        else:
            raise ValueError('Improper coordinates format')

    @classmethod
    def from_ase(cls, atoms_object):
        """
        Generate ``BathCell`` instance from ``ase.Atoms`` object of Atomic Simulations Environment (ASE) package.

        Args:
            atoms_object (Atoms): Atoms object, used to generate new ``BathCell`` instance.

        Returns:
            BathCell: New instance of the ``BathCell`` with atoms read from ``ase.Atoms``.
        """

        spin_cell = cls(cell=atoms_object.cell[:].T)
        positions = atoms_object.get_scaled_positions(wrap=True)
        symbols = atoms_object.get_chemical_symbols()

        zdr = np.linalg.inv(spin_cell.cell) @ np.array([0, 0, 1])

        zdr = zdr / np.linalg.norm(zdr)

        spin_cell._zdir = zdr

        for s in symbols:
            spin_cell.atoms[s] = []

        for sym, pos in zip(symbols, positions):
            spin_cell.atoms[sym].append(pos)

        return spin_cell

    def __repr__(self):
        m = f"{type(self).__name__} containing:\n"
        na = 0
        for k in self.atoms:
            na += 1
            m += f"{len(self.atoms[k])} positions for {k}"
            try:
                im = ' with'
                for ik in self.isotopes[k]:
                    im += f" {ik}: {self.isotopes[k][ik]}"
                im += '.\n'
            except KeyError:
                im = ".\n"
            m += im
        if not na:
            m += "No atomic positions.\n"

        m += f"\nCell:\n{self.cell}\n"
        m += f"\nz-direction: {self.zdir}\n"

        return m


_remove_digits = str.maketrans('', '', '+-^1234567890')


def random_bath(names, size, number=1000, density=None, types=None,
                density_units='cm-3', center=None,
                seed=None):
    r"""
    Generate random bath containing spins with names provided with argument ``name`` in the box of size ``size``.
    By default generates coordinates in range (-size/2; +size/2) but this behavior can be changed by providing
    ``center`` keyword.

    Examples:

        Generate 2000 :math:`^{13}\mathrm{C}` nuclear spins in the cubic box with the side of 100 angstrom::

            >>> atoms = random_bath('13C', 100, number=2000, seed=10)
            >>> print(atoms.size)
            2000
            >>> print(round(atoms.x.min()), round(atoms.x.max()))
            -50.0 50.0

        Generate electron spin bath with density :math:`10^{17} \mathrm{cm}^{-3}` in the cuboid box::

            >>> electrons = random_bath('e', [1e3, 2e3, 3e3], density=1e17,
            >>>                         density_units='cm-3', seed=10)
            >>> print(electrons.size, round(electrons.x.min()), round(electrons.x.max()))
            600 -494.0 500.0
            >>> print(electrons.types)
            SpinDict(e: (e, 0.5, -17608.59705))

    Args:
        names (str or array-like with length n): Name of the bath spin or array with the names of the bath spins,
        size (float or ndarray with shape (3,)): Size of the box. If float is given,
            assumes 3D cube with the edge = ``size``. Otherwise the size specifies the dimensions of the box.
            Dimensionality is controlled by setting entries of the size array to 0.

        number (int or array-like with length n): Number of the bath spins in the box
            or array with the numbers of the bath spins. Has to have the same length as the ``name`` array.

        density (float or array-like with length n): Concentration of the bath spin
            or array with the concentrations. Has to have the same length as the ``name`` array.

        types (SpinDict): Dictionary with SpinTypes or input to create one.

        density_units (str): If number of spins provided as density, defines units.
            Values are accepted in the format ``m``, or ``m^x`` or ``m-x`` where m is the length unit,
            x is dimensionality of the bath (e.g. x = 1 for 1D, 2 for 2D etc).
            If only ``m`` is provided the dimensions are inferred from ``size`` argument.
            Accepted length units:

                * ``m`` meters;
                * ``cm`` centimeters;
                * ``a`` angstroms.

        center (ndarray with shape (3,)): Coordinates of the (0, 0, 0) point of the final coordinate system
            in the initial coordinates. Default is ``size / 2`` - center is in the middle of the box.
        seed (int): Seed for random number generator.
    Returns:
        BathArray with shape (np.prod(number),): Array of the bath spins with random positions.
    """
    size = np.asarray(size)
    unit_conversion = {'a': 1, 'cm': 1e-8, 'm': 1e-10}
    name = np.asarray(names)

    if size.size == 1:
        size = np.array([size, size, size])

    elif size.size > 3:
        raise RuntimeError('Wrong size format')

    if center is None:
        center = size / 2

    if density is not None:
        du = density_units.lower().translate(_remove_digits)
        power = re.findall(r'\d+', density_units.lower())

        sc = np.count_nonzero(size != 0)

        if power:
            powa = int(power[0])
            if sc != powa:
                warnings.warn(f'size dimensions {sc} do not agree with density units {density_units}',
                              stacklevel=2)
        else:
            powa = sc

        density = np.asarray(density) * unit_conversion[du] ** powa

        number = np.rint(density * np.prod(size[size != 0])).astype(np.int32)

    else:
        number = np.asarray(number, dtype=np.int32)

    total_number = np.sum(number)
    spins = BathArray((total_number,), types=types)

    if name.shape:
        counter = 0
        for n, no in zip(name, number):
            spins.N[counter:counter + no] = n
            counter += no

    else:
        spins.N = name

    # Generate the coordinates
    generator = np.random.default_rng(seed=seed)

    spins.xyz = generator.random(spins.xyz.shape) * size - center
    return spins


def read_ase(atoms_object):
    """
    Generate ``BathCell`` instance from ``ase.Atoms`` object of Atomic Simulations Environment (ASE) package.

    Args:
        atoms_object (Atoms): Atoms object, used to generate new ``BathCell`` instance.

    Returns:
        BathCell: New instance of the ``BathCell`` with atoms read from ``ase.Atoms``.
    """

    return BathCell.from_ase(atoms_object)


def defect(cell, atoms, add=None, remove=None):
    """
    Generate a defect in the given supercell.

    The defect will be located in the unit cell, located roughly
    in the middle of the supercell, generated by ``BathCell``, such that (0, 0, 0) of cartesian reference frame
    is located at (0, 0, 0) position of this unit cell.

    Args:
        cell (ndarray with shape (3, 3)): parameters of the unit cell.

        atoms (BathArray): Array of spins in the supercell.

        add (tuple or list of tuples):
            Add spin type(s) to the supercell at specified positions to create point defect.
            Each tuple contains name of the new isotope
            and its coordinates in the cell basis: ``(isotope_name, x_cell, y_cell, z_cell)``.

        remove (tuple or list of tuples):
            Remove lattice site from the supercell at specified position to create point defect.
            Each tuple contains name of the atom to remove
            and its coordinates in the cell basis: ``(atom_name, x_cell, y_cell, z_cell)``.

    Returns:
        BathArray: Array of spins with the defect added.
    """

    defective_atoms = atoms.copy()
    dt = atoms.dtype

    where = None
    if remove is not None and isinstance(remove[0], str):
        name = remove[0]
        position_cc = np.asarray(remove[1])  # Given in the cell coordinates

        position = cell @ position_cc

        offsets = np.linalg.norm((atoms['xyz'] - position), axis=1)
        where = np.logical_and(np.core.defchararray.find(atoms['N'], name) != -1,
                               offsets <= err_range)

    elif remove is not None:
        where = np.zeros(defective_atoms.shape, dtype=bool)

        for removals in remove:
            name = removals[0]
            position_cc = np.asarray(removals[1])  # Given in the cell coordinates

            position = cell @ position_cc
            # print(name, position)
            # print(np.core.defchararray.find(bath['N'], name) != -1)
            offsets = np.linalg.norm((atoms['xyz'] - position), axis=1)
            # print(offsets <= err_range)
            where += np.logical_and(np.core.defchararray.find(atoms['N'], name) != -1,
                                    offsets <= err_range)
    if np.count_nonzero(where):
        # print('I see {} removals'.format(np.count_nonzero(where)))
        # print('Removing: \n', bath[where])

        defective_atoms = defective_atoms[~where]

    if add is not None and isinstance(add[0], str):
        name = add[0]
        position_cc = np.asarray(add[1])  # Given in the cell coordinates

        position = cell @ position_cc

        newentry = np.array((name, position), dtype=dt)

        # print('Adding: \n', newentry)
        defective_atoms = np.append(defective_atoms, newentry)

    elif add is not None:
        newlist = []

        for addition in add:
            name = addition[0]
            position_cc = np.asarray(addition[1])  # Given in the cell coordinates

            position = cell @ position_cc

            newentry = np.array((name, position), dtype=dt)

            newlist.append(newentry)

        # print('Adding: \n', np.asarray(newlist))

        defective_atoms = np.append(defective_atoms, newlist)

    return defective_atoms
